import tempfile
import os
import json
import yaml
import sys
import colorama
import git
import functools
import unittest
import random
import string
import re

import chisubmit.client.session as session
from click.testing import CliRunner
from chisubmit.cli import chisubmit_cmd
from dateutil.parser import parse
from chisubmit.client.session import BadRequestError

colorama.init()

class ChisubmitTestClient(object):
    
    API_PREFIX = "/api/v0/"
    
    def __init__(self, app, user_id, api_key):
        self.user_id = user_id
        self.api_key = api_key
        
        self.headers = {'content-type': 'application/json'}
        
        if self.api_key is not None:
            self.headers["CHISUBMIT-API-KEY"] = self.api_key
            
        self.test_client = session.connect_test(app, api_key)

    def get(self, resource):
        return self.test_client.get(self.API_PREFIX + resource, headers = self.headers)

    def post(self, resource, data):
        if isinstance(data, dict):
            datastr = json.dumps(data)
        elif isinstance(data, basestring):
            datastr = data
        return self.test_client.post(self.API_PREFIX + resource, data = datastr, headers = self.headers)

    def put(self, resource, data):
        if isinstance(data, dict):
            datastr = json.dumps(data)
        elif isinstance(data, basestring):
            datastr = data
        return self.test_client.put(self.API_PREFIX + resource, data = datastr, headers = self.headers)


def cli_test(func=None, isolated_filesystem = True):
    if func is None:
        return functools.partial(cli_test, isolated_filesystem = isolated_filesystem)
    def new_func(self, *args, **kwargs):
        runner = CliRunner()
        
        try:
            if isolated_filesystem:
                with runner.isolated_filesystem():
                    func(self, runner, *args, **kwargs)
            else:
                func(self, runner, *args, **kwargs)
        except BadRequestError, bre:
            bre.print_errors()
            raise
    return new_func


class ChisubmitCLITestClient(object):
    
    def __init__(self, user_id, api_key, runner, course = None,
                 git_credentials = {}, verbose = False):
        self.user_id = user_id
        self.home_dir = "test-fs/home/%s" % self.user_id
        self.conf_dir = "%s/.chisubmit" % (self.home_dir)
        self.conf_file = self.conf_dir + "/chisubmit.conf"
        self.runner = runner
        self.verbose = verbose
        self.course = course
        
        git_credentials.update({"Testing" : "testing-credentials"})

        os.makedirs(self.home_dir)
        os.mkdir(self.conf_dir)
        with open(self.conf_file, 'w') as f:
            conf = {"api-url": "NONE",
                    "api-key": api_key,
                    "git-credentials":
                        git_credentials
                    }
            yaml.safe_dump(conf, f, default_flow_style=False)   
            
    def run(self, subcommands, params = [], course = None, cmd=chisubmit_cmd, catch_exceptions=False, cmd_input = None):
        chisubmit_args =  ['--testing', '--dir', self.conf_dir, '--conf', self.conf_file]
        
        if course is not None:
            chisubmit_args += ['--course', course]
        elif self.course is not None:
            chisubmit_args += ['--course', self.course]
        
        if subcommands is None:
            cmd_args = []
        else:
            cmd_args = subcommands.split()

        cmd_args += params
        
        if self.verbose:
            l = []
            for ca in cmd_args:
                if " " in ca:
                    l.append('"%s"' % ca)
                else:
                    l.append(ca)
            s = ""
            s+= colorama.Style.BRIGHT + colorama.Fore.BLUE
            s+= "%s$" % self.user_id
            s+= colorama.Fore.WHITE
            s+= " chisubmit %s" % " ".join(l)
            s+= colorama.Style.RESET_ALL
            print s
        
        result = self.runner.invoke(cmd, chisubmit_args + cmd_args, catch_exceptions=catch_exceptions, input=cmd_input)
        
        if self.verbose and len(result.output) > 0:
            sys.stdout.write(result.output)
            sys.stdout.flush()
            
        return result
    
    def create_local_git_repository(self, path):
        git_path = "%s/%s" % (self.home_dir, path)
                
        repo = git.Repo.init(git_path)

        return repo, git_path

    def get_local_git_repository(self, path):
        git_path = "%s/%s" % (self.home_dir, path)
                
        repo = git.Repo(git_path)

        return repo, git_path


class BaseChisubmitTestCase(unittest.TestCase):
    
    def __init__(self, *args, **kwargs):
        super(BaseChisubmitTestCase, self).__init__(*args, **kwargs)

        self.git_server_connstr = "server_type=Testing;local_path=./test-fs/server"
        self.git_staging_connstr = "server_type=Testing;local_path=./test-fs/staging"
        self.git_server_user = None
        self.git_staging_user = None
        self.git_api_keys = {}        
    
    @classmethod
    def setUpClass(cls):
        from chisubmit.backend.webapp.api import ChisubmitAPIServer

        cls.server = ChisubmitAPIServer(debug = True)
        cls.db_fd, cls.db_filename = tempfile.mkstemp()
        cls.server.connect_sqlite(cls.db_filename)
        
        password_length=random.randint(50,80)
        cls.auth_password = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(password_length))
        cls.server.set_auth_testing(cls.auth_password)
            
    @classmethod
    def tearDownClass(cls):
        os.close(cls.db_fd)
        os.unlink(cls.db_filename)
    
    def set_git_server_connstr(self, connstr):
        self.git_server_connstr = connstr

    def set_git_server_user(self, user):
        self.git_server_user = user

    def set_git_staging_connstr(self, connstr):
        self.git_staging_connstr = connstr

    def set_git_staging_user(self, user):
        self.git_staging_user = user
        
    def add_api_key(self, git_type, apikey):
        self.git_api_keys[git_type] = apikey        
    
    def get_admin_test_client(self):
        return ChisubmitTestClient(self.server.app, "admin", "admin")
    
    def get_test_client(self, user = None):
        if user is None:
            return ChisubmitTestClient(self.server.app, "anonymous", None)
        else:
            return ChisubmitTestClient(self.server.app, user["id"], user["api_key"])
        
    def assert_http_code(self, response, expected):
        self.assertEquals(response.status_code, expected, "Expected HTTP response code %i, got %i" % (expected, response.status_code))
        
    
class ChisubmitTestCase(BaseChisubmitTestCase):
        
    def setUp(self):
        self.server.init_db()
        self.server.create_admin(api_key="admin")
        
        if hasattr(self, "FIXTURE"):
            load_fixture(self.server.db, self.FIXTURE)
        
        
    def tearDown(self):
        self.server.db.session.remove()
        self.server.db.drop_all()
    

class ChisubmitMultiTestCase(BaseChisubmitTestCase):
        
    @classmethod
    def setUpClass(cls):
        super(ChisubmitMultiTestCase, cls).setUpClass()
        cls.server.init_db()
        cls.server.create_admin(api_key="admin")
        
        if hasattr(cls, "FIXTURE"):
            load_fixture(cls.server.db, cls.FIXTURE)
        

    @classmethod
    def tearDownClass(cls):
        cls.server.db.session.remove()
        cls.server.db.drop_all()
        super(ChisubmitMultiTestCase, cls).tearDownClass()
    
class ChisubmitFixtureTestCase(ChisubmitMultiTestCase):
    
    def test_get_courses(self):        
        for course in self.FIXTURE["courses"].values():
            for instructor in course["instructors"]:
                c = self.get_test_client(self.FIXTURE["users"][instructor])
                response = c.get("courses")
                self.assert_http_code(response, 200)
         
                expected_ncourses = len([c for c in self.FIXTURE["courses"].values()
                                         if instructor in c["instructors"]])
         
                data = json.loads(response.get_data())        
                self.assertIn("courses", data)
                self.assertEquals(len(data["courses"]), expected_ncourses)
                 
    def test_get_course(self):
        for course in self.FIXTURE["courses"].values():
            for instructor in course["instructors"]:
                c = self.get_test_client(self.FIXTURE["users"][instructor])
                response = c.get("courses/" + course["id"])
                self.assert_http_code(response, 200)
                data = json.loads(response.get_data())        
                self.assertIn("course", data)
                self.assertEquals(data["course"]["name"], course["name"])
                 
        for course1 in self.FIXTURE["courses"].values():
            for course2 in self.FIXTURE["courses"].values():
                if course1 != course2:
                    for instructor in course1["instructors"]:    
                        c = self.get_test_client(self.FIXTURE["users"][instructor])
                        response = c.get("courses/" + course2["id"])
                        self.assert_http_code(response, 404)
 
    
def load_fixture(db, fixture):
    from chisubmit.backend.webapp.api.users.models import User
    from chisubmit.backend.webapp.api.assignments.models import Assignment
    from chisubmit.backend.webapp.api.courses.models import Course,\
        CoursesInstructors, CoursesGraders, CoursesStudents
        
    user_objs = {}
    
    for u_id, user in fixture["users"].items():
        u = User(first_name=user["first_name"], 
                   last_name=user["last_name"], 
                   id=user["id"],
                   api_key=user["api_key"])
        
        user_objs[u_id] = u
        db.session.add(u)

    for c_id, course in fixture["courses"].items():
        c = Course(id = course["id"],
                   name = course["name"])
        
        db.session.add(c)
        
        if course.has_key("instructors"):
            for instructor in course["instructors"]:
                o = user_objs[instructor]
                db.session.add(CoursesInstructors(instructor_id = o.id, 
                                                  course_id     = c.id))
        
        if course.has_key("graders"):
            for grader in course["graders"]:
                o = user_objs[grader]
                db.session.add(CoursesGraders(grader_id = o.id, 
                                              course_id = c.id))

        if course.has_key("students"):
            for student in course["students"]:
                o = user_objs[student]
                db.session.add(CoursesStudents(student_id = o.id, 
                                               course_id = c.id))

        if course.has_key("assignments"):
            for assignment in course["assignments"].values():
                db.session.add(Assignment(id = assignment["id"], 
                                          name = assignment["name"],
                                          deadline = parse(assignment["deadline"]),
                                          course_id = c.id))
    
    db.session.commit()
        
    
class ChisubmitIntegrationTestCase(ChisubmitTestCase):
        
    def create_user(self, admin_runner, user_id):
        from chisubmit.backend.webapp.api.users.models import User

        fname = "f_" + user_id
        lname = "l_" + user_id
        email = user_id + "@example.org"
        result = admin_runner.run("admin user add", [user_id, fname, lname, email])
        self.assertEquals(result.exit_code, 0)
            
        user = User.from_id(user_id)
        self.assertIsNotNone(user)
        self.assertEquals(user.id, user_id)            
        self.assertEquals(user.first_name, fname)            
        self.assertEquals(user.last_name, lname)            
        self.assertEquals(user.email, email)      
        
        # TODO: We manually set the API key here, since the CLI
        # and the server are not completely set up for fetching
        # API keys yet. 
        user.api_key = user_id
        self.server.db.session.add(user)
        self.server.db.session.commit()

    def create_clients(self, runner, course_id, admin_id, instructor_ids, grader_ids, student_ids):
        admin = ChisubmitCLITestClient(admin_id, admin_id, runner, git_credentials=self.git_api_keys, verbose = True)
        
        instructors = []
        for instructor_id in instructor_ids:
            instructors.append(ChisubmitCLITestClient(instructor_id, instructor_id, runner, git_credentials=self.git_api_keys,  verbose = True, course = course_id))

        graders = []
        for grader_id in grader_ids:
            graders.append(ChisubmitCLITestClient(grader_id, grader_id, runner, git_credentials=self.git_api_keys,  verbose = True, course = course_id))

        students = []
        for student_id in student_ids:
            students.append(ChisubmitCLITestClient(student_id, student_id, runner, git_credentials=self.git_api_keys,  verbose = True, course = course_id))
                        
        return admin, instructors, graders, students    
            
    def create_users(self, admin, user_ids):
        for user_id in user_ids:
            self.create_user(admin, user_id)

    def create_course(self, admin, course_id, course_name):
        from chisubmit.backend.webapp.api.courses.models import Course

        result = admin.run("admin course add", [course_id, course_name])
        self.assertEquals(result.exit_code, 0)
        
        course = Course.from_id(course_id)
        self.assertIsNotNone(course)
        self.assertEquals(course.name, course_name)
        
        result = admin.run("admin course list")
        self.assertEquals(result.exit_code, 0)
        self.assertIn(course_id, result.output)
        self.assertIn(course_name, result.output)
        
        git_server_connstr = self.git_server_connstr
        git_staging_connstr = self.git_staging_connstr

        result = admin.run("admin course set-option %s git-server-connstr %s" % (course_id, git_server_connstr))
        self.assertEquals(result.exit_code, 0)

        result = admin.run("admin course set-option %s git-staging-connstr %s" % (course_id, git_staging_connstr))
        self.assertEquals(result.exit_code, 0)
        
        course = Course.from_id(course_id)
        self.assertIn("git-server-connstr", course.options)
        self.assertIn("git-staging-connstr", course.options)
        self.assertEquals(course.options["git-server-connstr"], git_server_connstr)
        self.assertEquals(course.options["git-staging-connstr"], git_staging_connstr)
        
        result = admin.run("admin course unsetup-repo %s" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        result = admin.run("admin course setup-repo %s" % (course_id))
        self.assertEquals(result.exit_code, 0)

        result = admin.run("admin course unsetup-repo %s --staging" % (course_id))
        self.assertEquals(result.exit_code, 0)

        result = admin.run("admin course setup-repo %s --staging" % (course_id))
        self.assertEquals(result.exit_code, 0)        
        
        
    def add_users_to_course(self, admin, course_id, instructors, graders, students):
        from chisubmit.backend.webapp.api.courses.models import CoursesStudents        
        
        for instructor in instructors:
            result = admin.run("admin course add-instructor %s %s" % (course_id, instructor.user_id))
            self.assertEquals(result.exit_code, 0)
        
        for grader in graders:
            result = admin.run("admin course add-grader %s %s" % (course_id, grader.user_id))
            self.assertEquals(result.exit_code, 0)
        
        for student in students:
            result = admin.run("admin course add-student %s %s" % (course_id, student.user_id))
            self.assertEquals(result.exit_code, 0)
                
        for instructor in instructors:
            if self.git_server_user is None:
                git_username = "git-" + instructors[0].user_id
            else:
                git_username = self.git_server_user
    
            result = instructors[0].run("instructor user set-repo-option", 
                                    ["instructor", instructors[0].user_id, "git-username", git_username])
            self.assertEquals(result.exit_code, 0)

        for grader in graders:
            if self.git_server_user is None:
                git_username = "git-" + graders[0].user_id
            else:
                git_username = self.git_server_user
    
            result = instructors[0].run("instructor user set-repo-option", 
                                        ["grader", graders[0].user_id, "git-username", git_username])
            self.assertEquals(result.exit_code, 0)
                
        result = admin.run("admin course update-repo-access", [course_id])
        self.assertEquals(result.exit_code, 0)
                
        result = admin.run("admin course update-repo-access", [course_id, "--staging"])
        self.assertEquals(result.exit_code, 0)
        
        for student in students:
            if self.git_server_user is None:
                git_username = "git-" + student.user_id
            else:
                git_username = self.git_server_user
                
            result = student.run("student course set-git-username", [git_username])
            self.assertEquals(result.exit_code, 0)
            
            course_student = CoursesStudents.query.filter_by(
                course_id=course_id).filter_by(
                student_id=student.user_id).first()
                
            self.assertIn("git-username", course_student.repo_info)
            self.assertEquals(course_student.repo_info["git-username"], git_username)


    def create_team_repos(self, admin, course_id, teams, students_team):
        
        team_git_paths = []
        team_git_repos = []
        team_commits = []
        
        for team, students in zip(teams, students_team):
            result = admin.run("admin course team-repo-remove", [course_id, team, "--ignore-non-existing"])
            self.assertEquals(result.exit_code, 0)
            result = admin.run("admin course team-repo-remove", ["--staging", course_id, team, "--ignore-non-existing"])
            self.assertEquals(result.exit_code, 0)

            result = admin.run("admin course team-repo-create", [course_id, team, "--public"])
            self.assertEquals(result.exit_code, 0)
            result = admin.run("admin course team-repo-create", ["--staging", course_id, team, "--public"])
            self.assertEquals(result.exit_code, 0)

        
            result = students[0].run("student team repo-check", [team])
            self.assertEquals(result.exit_code, 0)
            
            r = re.findall(r"^Repository URL: (.*)$", result.output, flags=re.MULTILINE)
            self.assertEquals(len(r), 1, "student team repo-check didn't produce the expected output")
            repo_url = r[0]

            team_git_repo, team_git_path = students[0].create_local_git_repository(team)
            team_remote = team_git_repo.create_remote("origin", repo_url)
            
            team_git_repos.append(team_git_repo)
            team_git_paths.append(team_git_path)

            commits = []
                        
            files = ["foo", "bar", "baz"]
            
            for f in files:
                open("%s/%s" % (team_git_path, f), "w").close()
                
            team_git_repo.index.add(files)
            team_git_repo.index.commit("First commit of %s" % team)
                
            commits.append(team_git_repo.heads.master.commit)
    
            team_remote.push("master")
            
            f = open("%s/foo" % (team_git_path), "w")
            f.write("Hello, team1!")
            f.close()
            
            team_git_repo.index.add(["foo"])
            team_git_repo.index.commit("Second commit of %s" % team)

            commits.append(team_git_repo.heads.master.commit)
            
            team_remote.push("master")
            
            team_commits.append(commits)
            
            
        for team1, students1 in zip(teams, students_team):
            for team2, students2 in zip(teams, students_team):
                if team1 != team2:
                    result = students1[0].run("student team repo-check", [team2])
                    self.assertNotEquals(result.exit_code, 0)
                    result = students2[0].run("student team repo-check", [team1])
                    self.assertNotEquals(result.exit_code, 0)
            
        return team_git_paths, team_git_repos, team_commits

            
    def register_team(self, student_clients, team_name, assignment_id, course_id):
        from chisubmit.backend.webapp.api.users.models import User
        from chisubmit.backend.webapp.api.teams.models import Team, StudentsTeams

        for s in student_clients:
            partners = [s2 for s2 in student_clients if s2!=s]
            partner_args = []
            for p in partners:
                partner_args += ["--partner", p.user_id]
        
            s.run("student assignment register", 
                  [ assignment_id, "--team-name", team_name] + partner_args)

            s.run("student team show", [team_name])
        
        
        students_in_team = [User.from_id(s.user_id) for s in student_clients]
        
        ts = Team.find_teams_with_students(course_id, students_in_team)
        
        self.assertGreaterEqual(len(ts), 1)
        self.assertIn(team_name, [t.id for t in ts])
        
        team = [t for t in ts if t.id == team_name]
        self.assertEqual(len(team), 1)
        t = team[0]
        self.assertEquals(t.id, team_name)
        self.assertListEqual(sorted([s.id for s in students_in_team]), 
                             sorted([s.id for s in t.students]))
        for st in t.students_teams:
            self.assertEquals(st.status, StudentsTeams.STATUS_CONFIRMED)      
    


from chisubmit.backend.webapp.api import db
from sqlalchemy import not_
from flask.json import JSONEncoder
from datetime import datetime
import pytz


class Serializable(db.Model):
    __abstract__ = True

    # Stores changes made to this model's attributes. Can be retrieved
    # with model.changes
    _changes = {}

    def __init__(self, **kwargs):
        kwargs['_force'] = True
        self._set_columns(**kwargs)

    def _set_columns(self, **kwargs):
        force = kwargs.get('_force')

        readonly = []
        if hasattr(self, 'readonly_fields'):
            readonly = self.readonly_fields
        if hasattr(self, 'hidden_fields'):
            readonly += self.hidden_fields

        readonly += [
            'id',
            'created',
            'updated',
            'modified',
            'created_at',
            'updated_at',
            'modified_at',
        ]

        changes = {}

        columns = self.__table__.columns.keys()
        relationships = self.__mapper__.relationships.keys()

        for key in columns:
            allowed = True if force or key not in readonly else False
            exists = True if key in kwargs else False
            if allowed and exists:
                val = getattr(self, key)
                if val != kwargs[key]:
                    changes[key] = {'old': val, 'new': kwargs[key]}
                    setattr(self, key, kwargs[key])

        for rel in relationships:
            allowed = True if force or rel not in readonly else False
            exists = True if rel in kwargs else False
            if allowed and exists:
                is_list = self.__mapper__.relationships[rel].uselist
                if is_list:
                    valid_ids = []
                    query = getattr(self, rel)
                    cls = self.__mapper__.relationships[rel].argument()
                    for item in kwargs[rel]:
                        if 'id' in item and query.filter_by(
                                id=item['id']).limit(1).count() == 1:
                            obj = cls.query.filter_by(id=item['id']).first()
                            col_changes = obj.set_columns(**item)
                            if col_changes:
                                col_changes['id'] = str(item['id'])
                                if rel in changes:
                                    changes[rel].append(col_changes)
                                else:
                                    changes.update({rel: [col_changes]})
                            valid_ids.append(str(item['id']))
                        else:
                            col = cls()
                            col_changes = col.set_columns(**item)
                            query.append(col)
                            db.session.flush()
                            if col_changes:
                                col_changes['id'] = str(col.id)
                                if rel in changes:
                                    changes[rel].append(col_changes)
                                else:
                                    changes.update({rel: [col_changes]})
                            valid_ids.append(str(col.id))

                    # delete related rows that were not in kwargs[rel]
                    for item in query.filter(
                            not_(cls.id.in_(valid_ids))).all():
                        col_changes = {
                            'id': str(item.id),
                            'deleted': True,
                        }
                        if rel in changes:
                            changes[rel].append(col_changes)
                        else:
                            changes.update({rel: [col_changes]})
                        db.session.delete(item)

                else:
                    val = getattr(self, rel)
                    if self.__mapper__.relationships[rel].query_class:
                        if val is not None:
                            col_changes = val.set_columns(**kwargs[rel])
                            if col_changes:
                                changes.update({rel: col_changes})
                    else:
                        if val != kwargs[rel]:
                            setattr(self, rel, kwargs[rel])
                            changes[rel] = {'old': val, 'new': kwargs[rel]}

        return changes

    def set_columns(self, **kwargs):
        self._changes = self._set_columns(**kwargs)
        if 'modified' in self.__table__.columns:
            self.modified = datetime.utcnow()
        if 'updated' in self.__table__.columns:
            self.updated = datetime.utcnow()
        if 'modified_at' in self.__table__.columns:
            self.modified_at = datetime.utcnow()
        if 'updated_at' in self.__table__.columns:
            self.updated_at = datetime.utcnow()
        return self._changes

    @property
    def changes(self):
        return self._changes

    def reset_changes(self):
        self._changes = {}

    def to_dict(self, show=None, hide=None, path=None, show_all=None):
        """ Return a dictionary representation of this model.
        """

        if not show:
            show = []
        if not hide:
            hide = []
        hidden = []
        if hasattr(self, 'hidden_fields'):
            hidden = self.hidden_fields
        default = []
        if hasattr(self, 'default_fields'):
            default = self.default_fields

        ret_data = {}

        if not path:
            path = self.__tablename__.lower()

            def prepend_path(item):
                item = item.lower()
                if item.split('.', 1)[0] == path:
                    return item
                if len(item) == 0:
                    return item
                if item[0] != '.':
                    item = '.%s' % item
                item = '%s%s' % (path, item)
                return item
            show[:] = [prepend_path(x) for x in show]
            hide[:] = [prepend_path(x) for x in hide]

        columns = self.__table__.columns.keys()
        relationships = self.__mapper__.relationships.keys()
        properties = dir(self)

        for key in columns:
            check = '%s.%s' % (path, key)
            if check in hide or key in hidden:
                continue
            if show_all or key is 'id' or check in show or key in default:
                ret_data[key] = getattr(self, key)

        for key in relationships:
            check = '%s.%s' % (path, key)
            if check in hide or key in hidden:
                continue
            if show_all or check in show or key in default:
                is_list = self.__mapper__.relationships[key].uselist
                if is_list:
                    ret_data[key] = []
                    for item in getattr(self, key):
                        ret_data[key].append(item.to_dict(
                            show=show,
                            hide=hide,
                            path=('%s.%s' % (path, key.lower())),
                            show_all=show_all,
                        ))
                else:
                    if self.__mapper__.relationships[key].query_class:
                        child = getattr(self, key)
                        if child is None:
                            ret_data[key] = None
                        else:
                            ret_data[key] = getattr(self, key).to_dict(
                                show=show,
                                hide=hide,
                                path=('%s.%s' % (path, key.lower())),
                                show_all=show_all,
                            )
                    else:
                        ret_data[key] = getattr(self, key)

        for key in list(set(properties) - set(columns) - set(relationships)):
            if key.startswith('_'):
                continue
            check = '%s.%s' % (path, key)
            if check in hide or key in hidden:
                continue
            if show_all or check in show or key in default:
                val = getattr(self, key)
                ret_data[key] = []
                if hasattr(val, '__iter__'):
                    for item in val:
                        if hasattr(item, 'to_dict') and callable(item.to_dict):
                            ret_data[key].append(item.to_dict(
                                show=show,
                                hide=hide,
                                path=('%s.%s' % (path, key.lower())),
                                show_all=show_all,
                            ))
                else:
                    if hasattr(val, 'to_dict') and callable(val.to_dict):
                        ret_data[key] = getattr(self, key).to_dict(
                            show=show,
                            hide=hide,
                            path=('%s.%s' % (path, key.lower())),
                            show_all=show_all,
                        )

        return ret_data




#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import click
from chisubmit.common import ChisubmitException, CHISUBMIT_FAIL,\
    handle_unexpected_exception, CHISUBMIT_SUCCESS
import sys
from pprint import pprint
from requests.exceptions import HTTPError, ConnectionError
from chisubmit.cli.admin import admin
from chisubmit.cli.instructor import instructor
from chisubmit.cli.student import student
from chisubmit.client.session import BadRequestError
from chisubmit.cli.grader import grader
import getpass
from chisubmit.client.user import User
from docutils.utils.math.math2html import URL
config = None

import chisubmit.common.log as log
from chisubmit.config import Config
from chisubmit import RELEASE
from chisubmit.client import session

SUBCOMMANDS_NO_COURSE = [('course','create')]
SUBCOMMANDS_DONT_SAVE = ['course-create', 'course-install', 'course-generate-distributable', 'gh-token-create', 'shell']

VERBOSE = False
DEBUG = False 

@click.group(name="chisubmit")
@click.option('--conf', type=str, default=None)
@click.option('--dir', type=str, default=None)
@click.option('--course', type=str, default=None)
@click.option('--verbose', '-v', is_flag=True)
@click.option('--debug', is_flag=True)
@click.option('--testing', is_flag=True)
@click.version_option(version=RELEASE)
@click.pass_context
def chisubmit_cmd(ctx, conf, dir, course, verbose, debug, testing):
    global VERBOSE, DEBUG
    
    VERBOSE = verbose
    DEBUG = debug
    
    ctx.obj = {}

    config = Config(dir, conf)
    log.init_logging(verbose, debug)

    if not config['api-key']:
        raise click.BadParameter("Sorry, can't find your chisubmit api token")

    if testing:
        from chisubmit.backend.webapp.api import app 
        session.connect_test(app, access_token = config['api-key'])
    else:
        session.connect(config['api-url'], config['api-key'])

    if course:
        course_specified = True
        course_id = course
    else:
        course_specified = False
        course_id = config['default-course']

    ctx.obj["course_specified"] = course_specified
    ctx.obj["course_id"] = course_id
    ctx.obj["config"] = config
    ctx.obj["verbose"] = verbose
    ctx.obj["debug"] = debug

    return CHISUBMIT_SUCCESS


def chisubmit_cmd_wrapper():
    cmd_wrapper(chisubmit_cmd)


def cmd_wrapper(cmd):
    try:
        cmd.main()
    except BadRequestError, bre:
        print
        print "ERROR: There was an error processing this request"
        print
        print "URL: %s" % bre.request.url
        print "HTTP method: %s" % bre.request.method
        print "Error(s):"
        for noun, errors in bre.errors:
            if len(errors) == 1:
                print "  - %s: %s" % (noun, errors)
            else:
                print "  - %s:" % noun
                for error in errors:
                    print "    - %s" % error
        print        
    except HTTPError, he:
        print "ERROR: chisubmit server returned an HTTP error"
        print
        print "URL: %s" % he.request.url
        print "HTTP method: %s" % he.request.method
        print "Status code: %i" % he.response.status_code
        print "Message: %s" % he.response.reason
        if DEBUG:
            print
            print "HTTP REQUEST"
            print "------------"
            print "%s %s" % (he.request.method, he.request.url)
            print
            for hname, hvalue in he.request.headers.items():
                print "%s: %s" % (hname, hvalue) 
            print
            if he.request.body is not None:
                print he.request.body
            print
            print "HTTP RESPONSE"
            print "-------------"
            for hname, hvalue in he.response.headers.items():
                print "%s: %s" % (hname, hvalue) 
            print
            print he.response._content
    except ConnectionError, ce:
        print "ERROR: Could not connect to chisubmit server"
        print "URL: %s" % ce.request.url
    except ChisubmitException, ce:
        print "ERROR: %s" % ce.message
        if DEBUG:
            ce.print_exception()
        sys.exit(CHISUBMIT_FAIL)
    except Exception, e:
        handle_unexpected_exception()

chisubmit_cmd.add_command(admin)
chisubmit_cmd.add_command(instructor)
chisubmit_cmd.add_command(student)
chisubmit_cmd.add_command(grader)


@click.command(name="chisubmit-get-credentials")
@click.option('--conf', type=str, default=None)
@click.option('--dir', type=str, default=None)
@click.option('--verbose', is_flag=True)
@click.option('--debug', is_flag=True)
@click.option('--user', prompt='Enter your chisubmit username')
@click.option('--password', prompt='Enter your chisubmit password', hide_input=True)
@click.option('--url', type=str)
@click.option('--no-save', is_flag=True)
@click.option('--reset', is_flag=True)
@click.option('--testing', is_flag=True)
@click.pass_context
def chisubmit_get_credentials_cmd(ctx, conf, dir, verbose, debug, user, password, url, no_save, reset, testing):
    global VERBOSE, DEBUG
    
    VERBOSE = verbose
    DEBUG = debug
        
    config = Config(dir, conf)

    server_url = None
    if testing:
        from chisubmit.backend.webapp.api import app 
        session.connect_test(app)
    else:
        if config['api-url'] is None and url is None:
            print "No server URL specified. Please add it to your chisubmit.conf file"
            print "or use the --url option"
            ctx.exit(CHISUBMIT_FAIL)
            
        if url is not None:
            server_url = url
        else:
            server_url = config['api-url']
            
        session.connect(server_url)
            
    try:
        token, exists_prior, is_new = User.get_token(user, password, reset)
    except HTTPError, he:
        if he.response.status_code == 401:
            print "ERROR: Incorrect username/password"
            ctx.exit(CHISUBMIT_FAIL)
        else:
            raise    
    except ConnectionError, ce:
        print "ERROR: Could not connect to chisubmit server"
        print "URL: %s" % server_url
        ctx.exit(CHISUBMIT_FAIL) 

    if token:
        config['api-key'] = token
        if config['api-url'] is None and not no_save:
            config['api-url'] = url

        if not no_save:
            config.save()

        if is_new:
            ttype = "NEW"
        else:
            ttype = "EXISTING"
        
        click.echo("")
        click.echo("Your %s chisubmit access token is: %s" % (ttype, token))
        if not no_save:
            click.echo("")
            click.echo("The token has been stored in your chisubmit configuration file.")
            click.echo("You should now be able to use the chisubmit commands.")
        if exists_prior and is_new:
            click.echo("")
            click.echo("Your previous chisubmit access token has been cancelled.")
            click.echo("Make sure you run chisubmit-get-credentials on any other")
            click.echo("machines where you are using chisubmit.")
        click.echo("")
    else:
        click.echo("Unable to create token. Incorrect username/password.")

    return CHISUBMIT_SUCCESS

def chisubmit_get_credentials_cmd_wrapper():
    cmd_wrapper(chisubmit_get_credentials_cmd)


from chisubmit.cli.server import server_start, server_initdb

@click.group()
@click.option('--conf', type=str, default=None)
@click.option('--dir', type=str, default=None)
@click.option('--verbose', '-v', is_flag=True)
@click.option('--debug', is_flag=True)
@click.version_option(version=RELEASE)
@click.pass_context
def chisubmit_server_cmd(ctx, conf, dir, verbose, debug):
    ctx.obj = {}

    config = Config(dir, conf)
    log.init_logging(verbose, debug)
    
    ctx.obj["config"] = config
        
    return CHISUBMIT_SUCCESS


chisubmit_server_cmd.add_command(server_start)
chisubmit_server_cmd.add_command(server_initdb)




import click
from chisubmit.cli.common import pass_course, DATETIME
from chisubmit.client.assignment import Assignment, GradeComponent
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from dateutil.parser import parse
from chisubmit.common.utils import convert_datetime_to_utc,\
    convert_datetime_to_local
import operator
from chisubmit.cli.shared.course import shared_course_list
from chisubmit.cli.shared.assignment import shared_assignment_list


@click.group(name="assignment")
@click.pass_context
def instructor_assignment(ctx):
    pass


@click.command(name="add")
@click.argument('assignment_id', type=str)
@click.argument('name', type=str)
@click.argument('deadline', type=DATETIME)
@pass_course
@click.pass_context
def instructor_assignment_add(ctx, course, assignment_id, name, deadline):
    deadline = convert_datetime_to_utc(deadline)
    
    assignment = Assignment(id = assignment_id,
                            name = name,
                            deadline = deadline,
                            course_id = course.id)
    
    return CHISUBMIT_SUCCESS






@click.command(name="add-grade-component")
@click.argument('assignment_id', type=str)
@click.argument('grade_component_id', type=str)
@click.argument('description', type=str)
@click.argument('points', type=float)
@pass_course
@click.pass_context
def instructor_assignment_add_grade_component(ctx, course, assignment_id, grade_component_id, description, points):
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)

    grade_component = GradeComponent(id = grade_component_id, description = description, points=points)
    assignment.add_grade_component(grade_component)

    return CHISUBMIT_SUCCESS


@click.command(name="stats")
@click.argument('assignment_id', type=str)
@pass_course
@click.pass_context
def instructor_assignment_stats(ctx, course, assignment_id):
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)
        
    title = "Assignment '%s'" % (assignment.name)
    print title
    print "=" * len(title)
         
    student_dict = {s.user.id: s.user for s in course.students if not s.dropped}
    students = set(student_dict.keys())
    dropped = set([s.user.id for s in course.students if s.dropped])
    
    teams_unconfirmed = []
    nteams = 0
    nstudents = len(student_dict)
    nstudents_assignment = 0
    nsubmissions = 0
    
    for team in course.teams:
        if team.has_assignment(assignment.id):
            nteams += 1
            
            if team.has_submitted(assignment.id):
                nsubmissions += 1
            
            unconfirmed = False
            for student in team.students:
                student_id = student.user.id
                if student_id not in dropped:
                    try:
                        students.remove(student_id)
                        nstudents_assignment += 1
                        if student.status == 0: # Unconfirmed
                            unconfirmed = True
                    except KeyError, ke:
                        print "WARNING: Student %s seems to be in more than one team (offending team: %s)" % (student_id, team.id)
                
            if unconfirmed:
                teams_unconfirmed.append(team)
                
    assert(nstudents == len(students) + nstudents_assignment)                
               
    print 
    print "%i / %i students in %i teams have signed up for assignment %s" % (nstudents_assignment, nstudents, nteams, assignment.id)
    print
    print "%i / %i teams have submitted the assignment" % (nsubmissions, nteams)
    
    if ctx.obj["verbose"]:
        if len(students) > 0:
            print
            print "Students who have not yet signed up"
            print "-----------------------------------"
            not_signed_up = [student_dict[sid] for sid in students]
            not_signed_up.sort(key=operator.attrgetter("last_name"))
            for s in not_signed_up:
                print "%s, %s <%s>" % (s.last_name, s.first_name, s.email)
    
    return CHISUBMIT_SUCCESS


instructor_assignment.add_command(shared_assignment_list)

instructor_assignment.add_command(instructor_assignment_add)
instructor_assignment.add_command(instructor_assignment_add_grade_component)

instructor_assignment.add_command(instructor_assignment_stats)




from wtforms import Form
from wtforms.validators import InputRequired, Optional,\
    Length, Email
from wtforms.fields import StringField, BooleanField


class CreateUserInput(Form):
    id = StringField(validators=[Length(max=36), InputRequired()])
    first_name = StringField(validators=[Length(max=36), InputRequired()])
    last_name = StringField('last_name',
                            validators=[Length(max=36, min=1),
                                        InputRequired()])
    email = StringField(
        validators=[InputRequired(), Length(max=128, min=10), Email()])


class UpdateUserInput(Form):
    first_name = StringField(validators=[Length(max=36), InputRequired()])
    last_name = StringField('last_name',
                            validators=[Length(max=36, min=1),
                                        Optional()])
    email = StringField(
        validators=[InputRequired(), Length(max=128, min=10), Email()])

class GenerateAccessTokenInput(Form):
    reset = BooleanField(validators=[Optional()])



import click
from chisubmit.cli.student.assignment import student_assignment
from chisubmit.cli.student.team import student_team
from chisubmit.cli.student.course import student_course

@click.group()
@click.pass_context
def student(ctx):
    pass

student.add_command(student_assignment)
student.add_command(student_course)
student.add_command(student_team)

    







import click
from chisubmit.cli.common import pass_course, get_teams
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL,\
    ChisubmitException

import pprint
from chisubmit.cli.shared.team import shared_team_list, shared_team_show
import os
from chisubmit.repos.local import LocalGitRepo
from chisubmit.common.utils import create_connection
from git.exc import InvalidGitRepositoryError, GitCommandError
import operator

@click.group(name="team")
@click.pass_context
def instructor_team(ctx):
    pass


@click.command(name="search")
@click.option('--verbose', is_flag=True)
@click.argument('team_id', type=str)
@pass_course
@click.pass_context
def instructor_team_search(ctx, course, verbose, team_id):
    teams = course.search_team(team_id)

    pp = pprint.PrettyPrinter(indent=4, depth=6)

    for t in teams:
        tdict = dict(vars(t))
        if verbose:
            tdict["assignments"] = dict(tdict["assignments"])
            for p in tdict["assignments"]:
                tdict["assignments"][p] = vars(tdict["assignments"][p])

            tdict["students"] = [vars(s) for s in tdict["students"]]

        pp.pprint(tdict)

    return CHISUBMIT_SUCCESS

@click.command(name="pull-repos")
@click.argument('assignment_id', type=str)
@click.argument('directory', type=str)
@click.option('--only-ready-for-grading', is_flag=True)
@click.option('--reset', is_flag=True)
@click.option('--only', type=str)
@pass_course
@click.pass_context
def instructor_team_pull_repos(ctx, course, assignment_id, directory, only_ready_for_grading, reset, only):
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)

    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)

    teams = get_teams(course, assignment, only = only)

    directory = os.path.expanduser(directory)
    
    if not os.path.exists(directory):
        os.makedirs(directory)

    max_len = max([len(t.id) for t in teams])

    for team in sorted([t for t in teams if t.active], key=operator.attrgetter("id")):
        team_dir = "%s/%s" % (directory, team.id)
        team_git_url = conn.get_repository_git_url(course, team) 
        ta = team.get_assignment(assignment.id)

        if not team.has_assignment_ready_for_grading(assignment) and only_ready_for_grading:
            print "%-*s  SKIPPING (not ready for grading)" % (max_len, team.id)
            continue
        
        try:
            msg = ""
            if not os.path.exists(team_dir):
                r = LocalGitRepo.create_repo(team_dir, clone_from_url=team_git_url)
                msg = "Cloned repo"
            else:
                r = LocalGitRepo(team_dir)
                if reset:
                    r.fetch("origin")
                    r.reset_branch("origin", "master")
                    msg = "Reset to match origin/master" 
                else:
                    if r.repo.is_dirty():
                        print "%-*s  ERROR: Cannot pull. Local repository has unstaged changes." % (max_len, team.id)
                        continue
                    r.checkout_branch("master")
                    r.pull("origin", "master")
                    msg = "Pulled latest changes"
            if only_ready_for_grading:
                r.checkout_commit(ta.commit_sha)
                msg += " and checked out commit %s" % (ta.commit_sha)               
            print "%-*s  %s" % (max_len, team.id, msg)
        except ChisubmitException, ce:
            print "%-*s  ERROR: Could not checkout or pull master branch (%s)" % (max_len, team.id, ce.message)
        except GitCommandError, gce:
            print "%-*s  ERROR: Could not checkout or pull master branch" % (max_len, team.id)
            print gce
        except InvalidGitRepositoryError, igre:
            print "%-*s  ERROR: Directory %s exists but does not contain a valid git repository"  % (max_len, team.id, team_dir)
        except Exception, e:
            print "%-*s  ERROR: Unexpected exception when trying to checkout/pull" % (max_len, team.id)
            raise
    
        

    return CHISUBMIT_SUCCESS


@click.command(name="student-add")
@click.argument('team_id', type=str)
@click.argument('student_id', type=str)
@pass_course
@click.pass_context
def instructor_team_student_add(ctx, course, team_id, student_id):
    student = course.get_student(student_id)
    if student is None:
        print "Student %s does not exist" % student_id
        ctx.exit(CHISUBMIT_FAIL)

    team = course.get_team(team_id)
    if team is None:
        print "Team %s does not exist" % team_id
        ctx.exit(CHISUBMIT_FAIL)

    team.add_student(student)

    return CHISUBMIT_SUCCESS


@click.command(name="assignment-add")
@click.argument('team_id', type=str)
@click.argument('assignment_id', type=str)
@pass_course
@click.pass_context
def instructor_team_assignment_add(ctx, course, team_id, assignment_id):
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)

    team = course.get_team(team_id)
    if team is None:
        print "Team %s does not exist" % team_id
        ctx.exit(CHISUBMIT_FAIL)

    if team.assignments.has_key(assignment.id):
        print "Team %s has already been assigned assignment %s"  % (team.id, assignment.id)
        ctx.exit(CHISUBMIT_FAIL)

    team.add_assignment(assignment)

    return CHISUBMIT_SUCCESS

@click.command(name="set-active")
@click.argument('team_id', type=str)
@pass_course
@click.pass_context
def instructor_team_set_active(ctx, course, team_id):
    team = course.get_team(team_id)
    if team is None:
        print "Team %s does not exist" % team_id
        ctx.exit(CHISUBMIT_FAIL)

    team.active = True

    return CHISUBMIT_SUCCESS

@click.command(name="set-inactive")
@click.argument('team_id', type=str)
@pass_course
@click.pass_context
def instructor_team_set_inactive(ctx, course, team_id):
    team = course.get_team(team_id)
    if team is None:
        print "Team %s does not exist" % team_id
        ctx.exit(CHISUBMIT_FAIL)

    team.active = False

    return CHISUBMIT_SUCCESS

@click.command(name="set-attribute")
@click.argument('team_id', type=str)
@click.argument('name', type=str)
@click.argument('value', type=str)
@pass_course
@click.pass_context
def instructor_team_set_attribute(ctx, course, team_id, name, value):
    team = course.get_team(team_id)
    if team is None:
        print "Team %s does not exist" % team_id
        ctx.exit(CHISUBMIT_FAIL)

    team.set_extra(name, value)

    return CHISUBMIT_SUCCESS


instructor_team.add_command(shared_team_list)
instructor_team.add_command(shared_team_show)

instructor_team.add_command(instructor_team_search)
instructor_team.add_command(instructor_team_pull_repos)
instructor_team.add_command(instructor_team_student_add)
instructor_team.add_command(instructor_team_assignment_add)
instructor_team.add_command(instructor_team_set_active)
instructor_team.add_command(instructor_team_set_inactive)
instructor_team.add_command(instructor_team_set_attribute)





from flask.json import JSONEncoder
from datetime import datetime


class CustomJSONEncoder(JSONEncoder):

    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        return JSONEncoder.default(self, o)



import click
from chisubmit.cli.shared.course import shared_course_list,\
    shared_course_set_default, shared_course_get_git_credentials

@click.group(name="course")
@click.pass_context
def instructor_course(ctx):
    pass

instructor_course.add_command(shared_course_list)
instructor_course.add_command(shared_course_set_default)
instructor_course.add_command(shared_course_get_git_credentials)





#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

from chisubmit.client import CourseQualifiedApiObject, JSONObject
from chisubmit.client import session
from chisubmit.client.user import User
import json
from chisubmit.common.utils import convert_datetime_to_utc, get_datetime_now_utc
from dateutil import parser
from datetime import timedelta

class Grade(JSONObject):
    
    _api_attrs = ('points','grade_component_id')

class StudentTeam(JSONObject):
    
    _api_attrs = ('status',)
    _has_one = {'user': ('student', User)}

class AssignmentTeam(JSONObject):
    
    _api_attrs = ('extensions_used', 'commit_sha', 'submitted_at', 'assignment_id', 'penalties')
    _has_one = {'grader': ('grader', User)}
    _has_many = {'grades': 'grades'}
    
    def __init__(self, *args, **kwargs):
        if kwargs.has_key("submitted_at"):
            if kwargs["submitted_at"] is not None and not hasattr(kwargs["submitted_at"], 'isoformat'):
                kwargs["submitted_at"] = convert_datetime_to_utc(parser.parse(kwargs['submitted_at']))
                
        super(AssignmentTeam, self).__init__(*args, **kwargs)    
    
    def get_total_penalties(self):
        return sum([p for p in self.penalties.values()])
        
    def get_total_grade(self):
        points = sum([g.points for g in self.grades])
                
        return points + self.get_total_penalties()
    

    
class Team(CourseQualifiedApiObject):

    _api_attrs = ('id', 'active', 'course_id', 'extensions_available', 'extras')
    _primary_key = 'id'    
    _updatable_attributes = ('active',)
    _has_many = {'students': 'students_teams',
                 'assignments': 'assignments_teams',
                 }    
    
    def has_assignment(self, assignment_id):        
        return self.get_assignment(assignment_id) is not None    
    
    def get_assignment(self, assignment_id):
        ats = [at for at in self.assignments if at.assignment_id == assignment_id]
        
        if len(ats) == 1:
            return ats[0]
        else:
            return None        
        
    def has_submitted(self, assignment_id):
        assignment = self.get_assignment(assignment_id)
        if assignment is None:
            return False
        else:
            if assignment.submitted_at is not None:
                return True
            else:
                return False        
        
    def has_assignment_ready_for_grading(self, assignment, when=None):
        ta = self.get_assignment(assignment.id)
        
        if ta is None:
            return False
        
        if ta.submitted_at is None:
            return False
        
        if when is None:
            when = get_datetime_now_utc()
            
        deadline = assignment.deadline + timedelta(days=ta.extensions_used)
        
        if when > deadline:
            return True
        else:
            return False
    
    def add_student(self, student):
        attrs = {'team_id': self.id, 'student_id': student.id}
        data = json.dumps({'students': {'add': [attrs]}})
        session.put(self.url(), data=data)

    def add_assignment(self, assignment):
        attrs = {'team_id': self.id, 'assignment_id': assignment.id}
        data = json.dumps({'assignments': {'add': [attrs]}})
        session.put(self.url(), data=data)

    def set_assignment_grade(self, assignment_id, grade_component_id, points):
        attrs = {'assignment_id': assignment_id, 'grade_component_id': grade_component_id, 'points': points}
        data = json.dumps({'grades': {'add': [attrs]}})
        session.put(self.url(), data=data)

    def set_assignment_penalties(self, assignment_id, penalties):
        attrs = {'assignment_id': assignment_id, 'penalties': penalties}
        data = json.dumps({'grades': {'penalties': [attrs]}})
        session.put(self.url(), data=data)
        
    def set_assignment_grader(self, assignment_id, grader_id):
        attrs = {'assignment_id': assignment_id, 'grader_id': grader_id}
        data = json.dumps({'assignments': {'update': [attrs]}})
        session.put(self.url(), data=data)        
        
    def is_complete(self):
        return all([s.status == 1 for s in self.students])
        
    def get_confirmed_students(self):
        return [s for s in self.students if s.status == 1]
    
    def get_unconfirmed_students(self):
        return [s for s in self.students if s.status == 0]            
    
    def set_extra(self, name, value):
        data = {"extras": [ {'name': name, 'value': value} ] }
        data = json.dumps(data)
        session.put(self.url(), data=data) 
    




#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

CHISUBMIT_SUCCESS = 0
CHISUBMIT_FAIL = 1

import traceback

class ChisubmitException(Exception):
    def __init__(self, message, original_exception = None):
        Exception.__init__(self, message)
        self.original_exception = original_exception
        if original_exception is not None:
            self.traceback = traceback.format_exc()
        else:
            self.traceback = None

    def print_exception(self):
        print self.traceback
        
def handle_unexpected_exception():
    print "ERROR: Unexpected exception"
    print traceback.format_exc()
    exit(CHISUBMIT_FAIL)
        



import os.path

from chisubmit.common import ChisubmitException
from chisubmit.repos.local import LocalGitRepo
from chisubmit.common.utils import create_connection


class GradingGitRepo(object):
    def __init__(self, team, assignment, repo, repo_path, commit_sha):
        self.team = team
        self.assignment = assignment
        self.repo = repo
        self.repo_path = repo_path
        self.commit_sha = commit_sha

    @classmethod
    def get_grading_repo(cls, config, course, team, assignment):
        base_dir = config["directory"]
        
        ta = team.get_assignment(assignment.id)
        
        repo_path = cls.get_grading_repo_path(base_dir, course, team, assignment)
        if not os.path.exists(repo_path):
            return None
        else:
            repo = LocalGitRepo(repo_path)
            return cls(team, assignment, repo, repo_path, ta.commit_sha)

    @classmethod
    def create_grading_repo(cls, config, course, team, assignment):
        base_dir = config["directory"]
        
        conn_server = create_connection(course, config)
        if conn_server is None:
            raise ChisubmitException("Could not connect to git server")
        
        conn_staging = create_connection(course, config, staging = True)
        if conn_server is None:
            raise ChisubmitException("Could not connect to git staging server")        
        
        repo_path = cls.get_grading_repo_path(base_dir, course, team, assignment)
        server_url = conn_server.get_repository_git_url(course, team)
        staging_url = conn_staging.get_repository_git_url(course, team)

        ta = team.get_assignment(assignment.id)

        repo = LocalGitRepo.create_repo(repo_path, clone_from_url = server_url, remotes = [("staging", staging_url)])
        return cls(team, assignment, repo, repo_path, ta.commit_sha)

    def sync(self):
        self.repo.fetch("origin")
        self.repo.fetch("staging")
        self.repo.reset_branch("origin", "master")

    def create_grading_branch(self):
        branch_name = self.assignment.get_grading_branch_name()
        if self.repo.has_branch(branch_name):
            raise ChisubmitException("%s repository already has a %s branch" % (self.team.id, branch_name))

        if self.commit_sha is not None:
            commit = self.repo.get_commit(self.commit_sha)
            if commit is None:
                self.sync()
                commit = self.repo.get_commit(self.commit_sha)
                if commit is None:
                    raise ChisubmitException("%s repository does not have a commit %s" % (self.team.id, self.commit_sha))

            self.repo.create_branch(branch_name, self.commit_sha)
            self.repo.checkout_branch(branch_name)

    def has_grading_branch(self):
        branch_name = self.assignment.get_grading_branch_name()
        return self.repo.has_branch(branch_name)

    def has_grading_branch_staging(self):
        branch_name = self.assignment.get_grading_branch_name()
        return self.repo.has_remote_branch("staging", branch_name)

    def has_grading_branch_github(self):
        branch_name = self.assignment.get_grading_branch_name()
        return self.repo.has_remote_branch("origin", branch_name)

    def checkout_grading_branch(self):
        branch_name = self.assignment.get_grading_branch_name()
        if not self.repo.has_branch(branch_name):
            raise ChisubmitException("%s repository does not have a %s branch" % (self.team.id, branch_name))

        self.repo.checkout_branch(branch_name)

    def push_grading_branch_to_students(self):
        self.__push_grading_branch("origin")

    def push_grading_branch_to_staging(self):
        self.__push_grading_branch("staging", push_master = True)

    def pull_grading_branch_from_students(self):
        self.__pull_grading_branch("origin", pull_master = True)

    def pull_grading_branch_from_staging(self):
        self.__pull_grading_branch("staging")

    def set_grader_author(self):
        c = self.repo.repo.config_writer()

        c.set_value("user", "name", "chisubmit grader")
        c.set_value("user", "email", "do-not-email@example.org")

    def __push_grading_branch(self, remote_name, push_master = False):
        branch_name = self.assignment.get_grading_branch_name()

        if not self.repo.has_branch(branch_name):
            raise ChisubmitException("%s repository does not have a %s branch" % (self.team.id, branch_name))

        if push_master:
            self.repo.push(remote_name, "master")

        self.repo.push(remote_name, branch_name)

    def __pull_grading_branch(self, remote_name, pull_master = False):
        if pull_master:
            self.repo.checkout_branch("master")
            self.repo.pull(remote_name, "master")

        branch_name = self.assignment.get_grading_branch_name()
        if self.repo.has_branch(branch_name):
            self.repo.checkout_branch(branch_name)
            self.repo.pull(remote_name, branch_name)
        else:
            self.repo.fetch(remote_name, branch_name)
            self.repo.checkout_branch(branch_name)

    def commit(self, files, commit_message):
        return self.repo.commit(files, commit_message)

    def is_dirty(self):
        return self.repo.is_dirty()


    @staticmethod
    def get_grading_repo_path(base_dir, course, team, assignment):
        # TODO 18DEC14: This code could be a problem
        # The base_dir is passed from far away
        return "%s/repositories/%s/%s/%s" % (base_dir, course.id, assignment.id, team.id)
    
    



#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

from requests import exceptions, Session
from urlparse import urlparse
from pprint import pprint
import json
import sys
from requests.exceptions import HTTPError

# Disable the InsecurePlatformWarning warning
# This warning is thrown in versions of Python < 2.7.9, which includes pretty
# much all current Ubuntus.
# 
# Disabling this warning doesn't withhold information that the user could
# use to prevent a potentially insecure connection (as in the case of an
# untrusted certificate). The warning states that the version of Python and
# SSL used could lead to certain connections failing. So far, we haven't
# encountered this in any of our tests.
#
# More details here: https://urllib3.readthedocs.org/en/latest/security.html#insecureplatformwarning
try:
    from requests.packages.urllib3 import disable_warnings
    from requests.packages.urllib3.exceptions import InsecurePlatformWarning

    disable_warnings(category=InsecurePlatformWarning)
except ImportError, ie:
    pass

class BadRequestError(HTTPError):
    
    def __init__(self, method, url, errors, *args, **kwargs):
        self.method = method
        self.url = url
        self.errors = errors        
        super(BadRequestError, self).__init__(*args, **kwargs)
        
    def print_errors(self):
        print "HTTP request %s %s returned error code 400 (Bad Request)" %(self.method, self.url)
        if len(self.errors) == 0:
            print "No additional error messages returned."
        else:
            for noun, message in self.errors:
                print "%s: %s" % (noun, message)
    

endpoint = None
session = None
test_client = None
testing = False
test_headers = None

def connect(url, access_token = None):
    global endpoint, session, testing
    testing = False
    endpoint = url
    session = Session()
    session.headers['content-type'] = 'application/json'
    if access_token is not None:
        session.headers["CHISUBMIT-API-KEY"] = access_token
    
def connect_test(app, access_token = None):
    global endpoint, test_client, testing, test_headers
    endpoint = "/api/v0/"
    test_client = app.test_client()
    testing = True
    test_headers = {'content-type': 'application/json'}
    if access_token is not None:
        test_headers["CHISUBMIT-API-KEY"] = access_token

    return test_client

def raise_for_status(response):
    if testing:
        http_error_msg = ''
    
        if 400 <= response.status_code < 500:
            http_error_msg = '%s Client Error: %s' % (response.status_code, response.status)
    
        elif 500 <= response.status_code < 600:
            http_error_msg = '%s Server Error: %s' % (response.status_code, response.status)
    
        if http_error_msg:
            raise HTTPError(http_error_msg, response=response)
    else:
        response.raise_for_status()


def __process_response(response, method, url):
    if testing:
        response_text = response.get_data()
        response_exc = None
    else:
        response_text = response.text
        response_exc = response
        
    if response.status_code != 400:
        raise_for_status(response)
    
    try:
        if testing:
            data_json = json.loads(response.get_data())
        else:
            data_json = response.json()
    except ValueError, ve:
        if response.status_code == 400:
            raise BadRequestError(errors=[("unknown", "Unexpected error message: %s" % response_text)])
        else:
            raise ve
            
    if response.status_code == 400:
        error_result = []
        for noun, problem in data_json.get('errors', {}).items():
            error_result.append((noun, problem))
        raise BadRequestError(method = method,
                              url = url,
                              errors = error_result,
                              response = response_exc)

    return data_json


def get(resource, headers=None, **kwargs):
    url = endpoint + resource
    
    if testing:
        if headers is None:
            headers = test_headers
        else:
            headers.update(test_headers)
        response = test_client.get(url, headers=headers, **kwargs)
    else:    
        response = session.get(url, headers=headers, **kwargs)
        
    return __process_response(response, method = "GET", url = url) 

def post(resource, data, headers=None, **kwargs):
    url = endpoint + resource

    if testing:
        if headers is None:
            headers = test_headers
        else:
            headers.update(test_headers)        
        response = test_client.post(url, data=data, headers=headers, **kwargs)
    else:
        response = session.post(url, data, headers=headers, **kwargs)
    
    return __process_response(response, method = "POST", url = url) 

def put(resource, data, headers = None, **kwargs):
    url = endpoint + resource
    
    if testing:
        if headers is None:
            headers = test_headers
        else:
            headers.update(test_headers)        
        response = test_client.put(url, data=data, headers=headers, **kwargs)
    else:
        response = session.put(url, data, headers=headers, **kwargs)
    
    return __process_response(response, method = "PUT", url = url) 


def exists(obj):
    if isinstance(obj, (str, unicode)):
        obj_endpoint = obj
    else:
        obj_endpoint = obj.url()
    if testing:
        response = test_client.get(endpoint + obj_endpoint, headers=test_headers)
    else:
        response = session.get(endpoint + obj_endpoint)
    if response.status_code == 404:
        return False
    return True



import click
from chisubmit.cli.common import pass_course
from chisubmit.repos.factory import RemoteRepositoryConnectionFactory
from chisubmit.common import CHISUBMIT_FAIL, CHISUBMIT_SUCCESS

@click.group(name="user")
@click.pass_context
def instructor_user(ctx):
    pass

@click.command(name="set-repo-option")
@click.argument('user_type', type=click.Choice(["instructor", "grader", "student"]))
@click.argument('user_id', type=str)
@click.argument('option', type=str)
@click.argument('value', type=str)
@pass_course
@click.pass_context
def instructor_user_set_repo_option(ctx, course, user_type, user_id, option, value):
    if user_type == "instructor":
        course.set_instructor_repo_option(user_id, option, value)
    elif user_type == "grader":
        course.set_grader_repo_option(user_id, option, value)
    elif user_type == "student":
        course.set_student_repo_option(user_id, option, value)


@click.command(name="student-set-dropped")
@click.argument('student_id', type=str)
@pass_course
@click.pass_context
def instructor_student_set_dropped(ctx, course, student_id):
    student = course.get_student(student_id)
    if student is None:
        print "Student %s does not exist" % student_id
        ctx.exit(CHISUBMIT_FAIL)

    course.set_student_dropped(student_id)

    return CHISUBMIT_SUCCESS

@click.command(name="grader-add-conflict")
@click.argument('grader_id', type=str)
@click.argument('student_id', type=str)
@pass_course
@click.pass_context
def instructor_grader_add_conflict(ctx, course, grader_id, student_id):
    grader = course.get_grader(grader_id)
    if grader is None:
        print "Grader %s does not exist" % grader_id
        ctx.exit(CHISUBMIT_FAIL)

    student = course.get_student(student_id)
    if student is None:
        print "Student %s does not exist" % student_id
        ctx.exit(CHISUBMIT_FAIL)

    course.add_grader_conflict(grader, student_id)

    return CHISUBMIT_SUCCESS


instructor_user.add_command(instructor_user_set_repo_option)
instructor_user.add_command(instructor_student_set_dropped)
instructor_user.add_command(instructor_grader_add_conflict)




from chisubmit.tests.common import ChisubmitMultiTestCase, cli_test,\
    ChisubmitCLITestClient, ChisubmitTestCase
from chisubmit.tests.fixtures import users_and_courses
    
    
class CLIAdminCourse(ChisubmitTestCase):
            
    @cli_test
    def test_admin_course_add(self, runner):
        from chisubmit.backend.webapp.api.courses.models import Course

        admin = ChisubmitCLITestClient("admin", "admin", runner)
        
        course_id = u"cmsc12300"
        course_name = u"Foobarmentals of Foobar"
        
        result = admin.run("admin course add", [course_id, course_name])
        self.assertEquals(result.exit_code, 0)
        
        course = Course.from_id(course_id)
        self.assertIsNotNone(course)
        self.assertEquals(course.name, course_name)
 
        
class CLIAdminCourseFixture(ChisubmitMultiTestCase):
    
    FIXTURE = users_and_courses
        
    @cli_test
    def test_admin_course_list(self, runner):
        admin = ChisubmitCLITestClient("admin", "admin", runner)
        
        result = admin.run("admin course list")

        self.assertEquals(result.exit_code, 0)

        for course in self.FIXTURE["courses"].values():            
            self.assertIn(course["name"], result.output)
            


import click
from chisubmit.client.assignment import Assignment
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.common.utils import convert_datetime_to_local,\
    create_connection, get_datetime_now_utc, compute_extensions_needed
from dateutil.parser import parse
from chisubmit.cli.common import pass_course
from chisubmit.cli.shared.assignment import shared_assignment_list
from datetime import timedelta


@click.group(name="assignment")
@click.pass_context
def student_assignment(ctx):
    pass


@click.command(name="register")
@click.argument('assignment_id', type=str)
@click.option('--team-name', type=str)
@click.option('--partner', type=str, multiple=True)
@pass_course
@click.pass_context
def student_assignment_register(ctx, course, assignment_id, team_name, partner):
    a = Assignment.from_id(course.id, assignment_id)
        
    a.register(team_name = team_name,
               partners = partner)
    
    return CHISUBMIT_SUCCESS

@click.command(name="show-deadline")
@click.argument('assignment_id', type=str)
@click.option('--utc', is_flag=True)
@pass_course
@click.pass_context
def student_assignment_show_deadline(ctx, course, assignment_id, utc):
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)

    now_utc = get_datetime_now_utc()
    now_local = convert_datetime_to_local(now_utc)

    deadline_utc = assignment.deadline
    deadline_local = convert_datetime_to_local(deadline_utc)

    print assignment.name
    print
    if utc:
        print "      Now (Local): %s" % now_local.isoformat(" ")
        print " Deadline (Local): %s" % deadline_local.isoformat(" ")
        print
        print "        Now (UTC): %s" % now_utc.isoformat(" ")
        print "   Deadline (UTC): %s" % deadline_utc.isoformat(" ")
    else:
        print "      Now: %s" % now_local.isoformat(" ")
        print " Deadline: %s" % deadline_local.isoformat(" ")

    print

    extensions = compute_extensions_needed(now_utc, deadline_utc)

    if extensions == 0:
        diff = deadline_utc - now_utc
    else:
        diff = now_utc - deadline_utc

    days = diff.days
    hours = diff.seconds // 3600
    minutes = (diff.seconds//60)%60
    seconds = diff.seconds%60

    if extensions == 0:
        print "The deadline has not yet passed"
        print "You have %i days, %i hours, %i minutes, %i seconds left" % (days, hours, minutes, seconds)
    else:
        print "The deadline passed %i days, %i hours, %i minutes, %i seconds ago" % (days, hours, minutes, seconds)
        print "If you submit your assignment now, you will need to use %i extensions" % extensions

    return CHISUBMIT_SUCCESS


def print_commit(commit):
    print "      Commit: %s" % commit.sha
    print "        Date: %s" % commit.committed_date.isoformat(sep=" ")
    print "     Message: %s" % commit.message
    print "      Author: %s <%s>" % (commit.author_name, commit.author_email)        


@click.command(name="submit")
@click.argument('team_id', type=str)    
@click.argument('assignment_id', type=str)
@click.argument('commit_sha', type=str)
@click.option('--extensions', type=int, default=0)
@click.option('--force', is_flag=True)
@click.option('--yes', is_flag=True)
@pass_course
@click.pass_context  
def student_assignment_submit(ctx, course, team_id, assignment_id, commit_sha, extensions, force, yes):
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)
    
    team = course.get_team(team_id)
    if team is None:
        print "Team %s does not exist" % team_id
        ctx.exit(CHISUBMIT_FAIL)
    
    ta = team.get_assignment(assignment_id)
    if ta is None:
        print "Team %s is not registered for assignment %s" % (team_id, assignment_id)
        ctx.exit(CHISUBMIT_FAIL)
        
    if team.has_assignment_ready_for_grading(assignment):
        print "You cannot re-submit this assignment."
        print "You made a submission before the deadline, and the deadline has passed."

        ctx.exit(CHISUBMIT_FAIL)
        
    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)
    
    commit = conn.get_commit(course, team, commit_sha)
    
    if commit is None:
        print "Commit %s does not exist in repository" % commit_sha
        ctx.exit(CHISUBMIT_FAIL)

    response = assignment.submit(team_id, commit_sha, extensions, dry_run=True)

    success = response["success"]
    dry_run = response["dry_run"]

    deadline_utc = parse(response["submission"]["deadline"])
    
    submitted_at_utc = parse(response["submission"]["submitted_at"])
    extensions_needed = response["submission"]["extensions_needed"]
    extensions_requested = response["submission"]["extensions_requested"]    

    extensions_available_before = response["team"]["extensions_available_before"]
    extensions_available = response["team"]["extensions_available"]

    if response["prior_submission"]["submitted_at"] is not None:    
        prior_submitted_at_utc = parse(response["prior_submission"]["submitted_at"])
        prior_submitted_at_local = convert_datetime_to_local(prior_submitted_at_utc)
    prior_commit_sha = response["prior_submission"]["commit_sha"]
    prior_extensions_used = response["prior_submission"]["extensions_used"]

    deadline_local = convert_datetime_to_local(deadline_utc)
    submitted_at_local = convert_datetime_to_local(submitted_at_utc)
        
    if not success:
        if extensions_needed > extensions_available:
            msg1 = "You do not have enough extensions to submit this assignment."
            msg2 = "You would need %i extensions to submit this assignment at this " \
                   "time, but you only have %i left" % (extensions_needed, extensions_available)
        elif extensions_requested < extensions_needed:
            msg1 = "The number of extensions you have requested is insufficient."
            msg2 = "You need to request %s extensions." % extensions_needed
        elif extensions_requested > extensions_needed:
            msg1 = "The number of extensions you have requested is excessive."
            msg2 = "You only need to request %s extensions." % extensions_needed

        print
        print msg1
        print            
        print "     Deadline (UTC): %s" % deadline_utc.isoformat(sep=" ")
        print "          Now (UTC): %s" % submitted_at_utc.isoformat(sep=" ")
        print 
        print "   Deadline (Local): %s" % deadline_local.isoformat(sep=" ")
        print "        Now (Local): %s" % submitted_at_local.isoformat(sep=" ")
        print 
        print msg2 
        print

        ctx.exit(CHISUBMIT_FAIL)
    else:             
        if prior_commit_sha is not None:
            submission_commit = conn.get_commit(course, team, prior_commit_sha)
            
            if prior_commit_sha == commit_sha:
                print "You have already submitted assignment %s" % assignment.id
                print "You submitted the following commit on %s:" % prior_submitted_at_local
                print
                if submission_commit is None:
                    print "WARNING: Previously submitted commit '%s' is not in the repository!" % prior_commit_sha
                else:
                    print_commit(submission_commit)
                print
                print "You are trying to submit the same commit again (%s)" % prior_commit_sha
                print "If you want to re-submit, please specify a different commit"
                ctx.exit(CHISUBMIT_FAIL)
                
            if not force:
                print        
                print "You have already submitted assignment %s" % assignment.id
                print "You submitted the following commit on %s:" % prior_submitted_at_local
                print
                if submission_commit is None:
                    print "WARNING: Previously submitted commit '%s' is not in the repository!" % prior_commit_sha
                else:
                    print_commit(submission_commit)
                print
                print "If you want to submit again, please use the --force option"
                ctx.exit(CHISUBMIT_FAIL)
            else:
                print
                print "WARNING: You have already submitted assignment %s and you" % assignment.id 
                print "are about to overwrite the previous submission of the following commit:"
                print
                if submission_commit is None:
                    print "WARNING: Previously submitted commit '%s' is not in the repository!" % prior_commit_sha
                else:
                    print_commit(submission_commit)
                print
    
        if prior_commit_sha is not None and force:
            msg = "THE ABOVE SUBMISSION FOR %s (%s) WILL BE CANCELLED." % (assignment.id, assignment.name)
            
            print "!"*len(msg)
            print msg
            print "!"*len(msg)
            print
            print "If you continue, your submission for %s (%s)" % (assignment.id, assignment.name)
            print "will now point to the following commit:"                
        else:
            print "You are going to make a submission for %s (%s)." % (assignment.id, assignment.name)
            print "The commit you are submitting is the following:"
        print
        print_commit(commit)
        print
        print "PLEASE VERIFY THIS IS THE EXACT COMMIT YOU WANT TO SUBMIT"
        print
        print "Your team currently has %i extensions" % (extensions_available_before)
        print
        if prior_commit_sha is not None:
            print "You used %i extensions in your previous submission of this assignment." % prior_extensions_used
            print "and you are going to use %i additional extensions now." % (extensions_needed - prior_extensions_used)
        else:
            print "You are going to use %i extensions on this submission." % extensions_needed
        print
        print "You will have %i extensions left after this submission." % extensions_available
        print 
        print "Are you sure you want to continue? (y/n): ", 
        
        if not yes:
            yesno = raw_input()
        else:
            yesno = 'y'
            print 'y'
        
        if yesno in ('y', 'Y', 'yes', 'Yes', 'YES'):
            response = assignment.submit(team_id, commit_sha, extensions, dry_run=False)
              
            # TODO: Can't do this until GitLab supports updating tags
            #    
            # message = "Extensions: %i\n" % extensions_requested
            # if submission_tag is None:
            #     conn.create_submission_tag(course, team, tag_name, message, commit.sha)
            # else:
            #     conn.update_submission_tag(course, team, tag_name, message, commit.sha)
                
            print
            if response["success"]:
                print "Your submission has been completed."
            else:
                print "ERROR: Your submission was not completed."
            
        return CHISUBMIT_SUCCESS
    
    
@click.command(name="cancel-submit")
@click.argument('team_id', type=str)    
@click.argument('assignment_id', type=str)
@click.option('--yes', is_flag=True)
@pass_course
@click.pass_context  
def student_assignment_cancel_submit(ctx, course, team_id, assignment_id, yes):
    assignment = course.get_assignment(assignment_id)
    if assignment is None:
        print "Assignment %s does not exist" % assignment_id
        ctx.exit(CHISUBMIT_FAIL)
    
    team = course.get_team(team_id)
    if team is None:
        print "Team %s does not exist" % team_id
        ctx.exit(CHISUBMIT_FAIL)
        
    ta = team.get_assignment(assignment_id)
    if ta is None:
        print "Team %s is not registered for assignment %s" % (team_id, assignment_id)
        ctx.exit(CHISUBMIT_FAIL)
        
    if ta.submitted_at is None:
        print "Team %s has not made a submission for assignment %s," % (team_id, assignment_id)
        print "so there is nothing to cancel."
        ctx.exit(CHISUBMIT_FAIL)
        
    if ta.submitted_at is not None:
        now = get_datetime_now_utc()
        deadline = assignment.deadline + timedelta(days=ta.extensions_used)
        
        if now > deadline:
            print "You cannot cancel this submission."
            print "You made a submission before the deadline, and the deadline has passed."

            ctx.exit(CHISUBMIT_FAIL)        
            
    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        print "Could not connect to git server."
        ctx.exit(CHISUBMIT_FAIL)
        
    submission_commit = conn.get_commit(course, team, ta.commit_sha)
        
    print
    print "This is your existing submission for assignment %s:" % assignment.id 
    print
    if submission_commit is None:
        print "WARNING: Previously submitted commit '%s' is not in the repository!" % ta.commit_sha
    else:
        print_commit(submission_commit)
    print    

    print "Are you sure you want to cancel this submission? (y/n): ", 
    
    if not yes:
        yesno = raw_input()
    else:
        yesno = 'y'
        print 'y'
    
    if yesno in ('y', 'Y', 'yes', 'Yes', 'YES'):
        response = assignment.cancel(team_id)
        
        # TODO: Can't do this until GitLab supports updating tags
        #    
        # message = "Extensions: %i\n" % extensions_requested
        # if submission_tag is None:
        #     conn.create_submission_tag(course, team, tag_name, message, commit.sha)
        # else:
        #     conn.update_submission_tag(course, team, tag_name, message, commit.sha)
            
        print
        if response["success"]:
            print "Your submission has been cancelled."
        else:
            print "ERROR: Your submission was not cancelled."
        

student_assignment.add_command(shared_assignment_list)

student_assignment.add_command(student_assignment_register)
student_assignment.add_command(student_assignment_show_deadline)
student_assignment.add_command(student_assignment_submit)
student_assignment.add_command(student_assignment_cancel_submit)



from functools import wraps
from flask import request, abort, g
from chisubmit.backend.webapp.api.users.models import User


def require_apikey(view_function):
    @wraps(view_function)
    
    # the new, post-decoration function. Note *args and **kwargs here.
    def decorated_function(*args, **kwargs):
        api_key_param = request.args.get('api-key')
        api_key_header = request.headers.get("CHISUBMIT-API-KEY")
        
        if api_key_param is None and api_key_header is None:
            api_key = None
        else:
            if api_key_param is None:
                api_key = api_key_header
            else:
                api_key = api_key_param
            
        if api_key is not None:
            user = User.query.filter_by(api_key=api_key).first()
            
            if user is None:
                abort(401)
            else:
                g.user = user
                return view_function(*args, **kwargs)
        else:
            abort(401)
    return decorated_function


from wtforms import Form
from wtforms.validators import InputRequired, Optional, Length
from wtforms.fields import StringField, IntegerField,\
    FormField, FieldList, BooleanField
from wtforms.fields.core import FloatField


class CreateTeamInput(Form):
    id = StringField(validators=[Length(max=36), InputRequired()])
    course_id = StringField(validators=[Length(max=36), InputRequired()])


class LinkAssignmentInput(Form):
    team_id = StringField(InputRequired())
    assignment_id = StringField(InputRequired())


class LinkStudentInput(Form):
    team_id = StringField(InputRequired())
    student_id = StringField(InputRequired())

class UpdateAssignmentTeamInput(Form):
    assignment_id = StringField(InputRequired())
    grader_id = StringField(Optional())

class AddAssignmentsInput(Form):
    add = FieldList(FormField(LinkAssignmentInput))
    update = FieldList(FormField(UpdateAssignmentTeamInput))


class AddPenaltiesInput(Form):
    assignment_id = StringField(InputRequired())
    penalties = StringField(InputRequired())

class AddGradeInput(Form):
    assignment_id = StringField(InputRequired())
    grade_component_id = StringField(InputRequired())
    points = FloatField(InputRequired())


class AddGradesInput(Form):
    add = FieldList(FormField(AddGradeInput))
    penalties = FieldList(FormField(AddPenaltiesInput))

class AddStudentsInput(Form):
    add = FieldList(FormField(LinkStudentInput))

class SetOptionInput(Form):
    name = StringField(validators=[InputRequired()])
    value = StringField(validators=[InputRequired()])

class UpdateTeamInput(Form):
    students = FormField(AddStudentsInput)
    assignments = FormField(AddAssignmentsInput)
    grades = FormField(AddGradesInput)
    extras = FieldList(FormField(SetOptionInput))
    active = BooleanField(validators=[Optional()])





from wtforms import Field
from wtforms.widgets import TextInput
from dateutil.parser import parse


class ISODateTimeField(Field):
    widget = TextInput()

    def _value(self):
        if self.data:
            return self.data.isoformat()
        else:
            return u''

    def process_formdata(self, valuelist):
        if valuelist:
            # convert [string] datetime.isoformat() back to a datetime.
            # strptime ***CAN NOT*** and Python has no builtin. Really.
            self.data = parse(valuelist[0])
        else:
            # TODO 12DEC:14: is this impossible?
            # there's nothing valid to put in here
            self.data = ''



import click
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.common.utils import create_connection
from chisubmit.cli.common import pass_course
from chisubmit.cli.shared.team import shared_team_list, shared_team_show
import tempfile
from chisubmit.repos.local import LocalGitRepo


@click.group(name="team")
@click.pass_context
def student_team(ctx):
    pass


@click.command(name="repo-check")
@click.argument('team_id', type=str)
@pass_course
@click.pass_context
def student_repo_check(ctx, course, team_id):
    team = course.get_team(team_id)
    if team is None:
        print "Team %s does not exist or you do not have access to it" % team_id
        ctx.exit(CHISUBMIT_FAIL)
    
    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        ctx.exit(CHISUBMIT_FAIL)

    if not conn.exists_team_repository(course, team):
        print "The repository for '%s' does not exist or you do not have permission to access it." % team_id
        ctx.exit(CHISUBMIT_FAIL)

    # TODO: Check that the user actually has push access
    print "Your repository exists and you have access to it."
    http_url = conn.get_repository_http_url(course, team)
    if http_url is not None:
        print "Repository website: %s" % conn.get_repository_http_url(course, team)
    print "Repository URL: %s" % conn.get_repository_git_url(course, team)

    return CHISUBMIT_SUCCESS

@click.command(name="repo-pristine-clone")
@click.argument('team_id', type=str)
@pass_course
@click.pass_context
def student_repo_pristine_clone(ctx, course, team_id):
    team = course.get_team(team_id)
    if team is None:
        print "Team %s does not exist or you do not have access to it" % team_id
        ctx.exit(CHISUBMIT_FAIL)
    
    conn = create_connection(course, ctx.obj['config'])
    
    if conn is None:
        ctx.exit(CHISUBMIT_FAIL)

    if not conn.exists_team_repository(course, team):
        print "The repository for '%s' does not exist or you do not have permission to access it." % team_id
        ctx.exit(CHISUBMIT_FAIL)

    tempdir = tempfile.mkdtemp(prefix="%s-%s-" % (course.id, team.id))
    
    repo_url = conn.get_repository_git_url(course, team)
    
    try:
        LocalGitRepo.create_repo(tempdir, clone_from_url=repo_url)
    except Exception, e:
        print "Unable to create a clone of repository %s" % repo_url
        ctx.exit(CHISUBMIT_FAIL)
        
    print "A pristine clone of your repository has been created in %s" % tempdir    

    return CHISUBMIT_SUCCESS

student_team.add_command(shared_team_list) 
student_team.add_command(shared_team_show)

student_team.add_command(student_repo_check)  
student_team.add_command(student_repo_pristine_clone)



#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.
import re
import json
from chisubmit.client import session
from datetime import datetime
import sys
from requests.exceptions import HTTPError

def json_default(obj):
    if isinstance(obj, datetime):
        return obj.isoformat()


class AttrOverride(type):

    def __getattr__(self, name):
        if name == 'singularize':
            return re.sub('(?!^)([A-Z]+)', r'_\1', self.__class__.__name__).lower()
        if name == 'pluralize':
            return self.singularize + 's'
        elif name == 'identifier':
            return self.__class__.__name__.lower() + '_id'
        else:
            raise AttributeError()

    def __new__(meta_class, name, bases, classdict):
        new_cls = type.__new__(meta_class, name, bases, classdict)
        for b in bases:
            if hasattr(b, 'register_subclass'):
                b.register_subclass(new_cls)
        return new_cls


class JSONObject(object):

    __metaclass__ = AttrOverride
    _subclasses = []

    def __init__(self, *args, **kwargs):
        class_id = self.__class__.__name__.lower() + '_id'
        for attr in self._api_attrs:
            if attr not in kwargs:
                kwargs[attr] = None
                
            super(JSONObject, self).__setattr__(attr, kwargs[attr])
                
            # FIXME 16JULY14: "alias" the id attribute.
            if attr == class_id:
                super(JSONObject, self).__setattr__("id", kwargs[attr])
        

    def __getattr__(self, name):
        class_attrs = self.__class__.__dict__

        if name == 'singularize':
            if '_singularize' in class_attrs:
                return class_attrs['_singularize']
            return re.sub('(?!^)([A-Z]+)', r'_\1', self.__class__.__name__).lower()
        if name == 'pluralize':
            if '_pluralize' in class_attrs:
                return class_attrs['_pluralize']
            return self.singularize() + 's'
        elif name == 'identifier':
            if '_primary_key' in class_attrs:
                id_field = class_attrs['_primary_key']
            else:
                id_field = self.singularize() + '_id'
            return getattr(self, id_field, None)
        elif '_updatable_attributes' in class_attrs and name in class_attrs['_updatable_attributes']:
            if name in self._json:
                return self._json[name]
            else:
                return None
        elif '_has_one' in class_attrs and name in class_attrs['_has_one']:
            attr = class_attrs['_has_one'][name][0]
            if self._json[attr] is None:
                return None
            cls = class_attrs['_has_one'][name][1]
            if not cls in ApiObject._subclasses:
                raise NameError
            else:
                return cls.from_json(self._json[attr])
        elif '_has_many' in class_attrs and name in class_attrs['_has_many']:
            attr = class_attrs['_has_many'][name]
            results = []
            singular_name = "_".join( [re.sub('s$', '', subname) for subname in attr.split('_') ])
            cls = singular_name.title().translate(None, '_')
            for item in self._json[attr]:
                c = None
                for subclass in ApiObject._subclasses:
                    if cls == subclass.__name__:
                        c = subclass
                if not c:
                    raise NameError
                item_from_uri = \
                    c.from_json(item)
                results.append(item_from_uri)
            return results
        else:
            type(self.__class__).__getattr__(self.__class__, name)

    @classmethod
    def register_subclass(klass, cls):
        klass._subclasses.append(cls)

    @classmethod
    def pluralize(cls):
        return cls.singularize() + 's'

    @classmethod
    def singularize(cls):
        return re.sub('(?!^)([A-Z]+)', r'_\1', cls.__name__).lower()

    @classmethod
    def from_json(cls, data):
        obj = cls(**data)
        obj._json = data
        return obj
    
    def __repr__(self):
        representation = "<%s -- " % (self.__class__.__name__)
        attrs = []
        for attr in self._api_attrs:
            attrs.append("%s:%s" % (attr, repr(getattr(self, attr))))
        return representation + ', '.join(attrs) + '>'
    

class BaseApiObject(JSONObject):

    def __setattr__(self, name, value):
        class_attrs = self.__class__.__dict__
        class_name = self.__class__.__name__

        if '_updatable_attributes' in class_attrs and name in class_attrs['_updatable_attributes']:
            self._api_update(name, value)
            super(BaseApiObject, self).__setattr__(name, value)
        elif name.startswith("_"):
            super(BaseApiObject, self).__setattr__(name, value)
        else:
            raise AttributeError("%s.%s is not an updatable attribute" % (class_name, name))

    @classmethod
    def from_json(cls, data):
        obj = cls(backendSave = False, **data)
        obj._json = data
        return obj

    @classmethod
    def from_uri(cls, uri):
        try:
            result = session.get(uri)
            return cls.from_json(data=result[cls.singularize()])
        except HTTPError, he:
            if he.response.status_code == 404:
                return None
            else:
                raise            
            

    def __init__(self, backendSave=True, **kwargs):
        super(BaseApiObject, self).__init__(**kwargs)
                
        if backendSave:
            self.save()

    def _api_update(self, attr, value):
        data = json.dumps({attr: value}, default=json_default)
        result = session.put(self.url(), data=data)
        self._json = result

    def save(self):
        attributes = {'id': self.identifier}
        attributes.update({attr: getattr(self, attr) for attr in self._api_attrs})        
        data = json.dumps(attributes, default=json_default)
        session.post(self.url(with_id=False), data)

class ApiObject(BaseApiObject):

    def __init__(self, *args, **kwargs):
        super(ApiObject, self).__init__(*args, **kwargs)

    def url(self, with_id = True):
        i = self.identifier
                   
        if with_id:
            i = self.identifier
            id_suffix = "/%s" % i
        else:
            id_suffix = ""
            
        return '%s%s' % (self.pluralize(), id_suffix)
    
    @classmethod
    def from_id(cls, identifier):
        url = cls.pluralize() + '/%s' % identifier
        return cls.from_uri(url)
    
    @classmethod
    def all(cls):
        json = session.get(cls.pluralize())
        return [cls.from_json(o) for o in json[cls.pluralize()]]    


class CourseQualifiedApiObject(BaseApiObject):

    def __init__(self, *args, **kwargs):
        if "course_id" not in kwargs:
            raise AttributeError("CourseQualifiedApiObject constructor did not get a 'course_id' keyword argument") 
        super(CourseQualifiedApiObject, self).__init__(*args, **kwargs)
        
    def url(self, with_id = True):
        i = self.identifier
                
        url_prefix = 'courses/%s/' % self.course_id
            
        if with_id:
            i = self.identifier
            id_suffix = "/%s" % i
        else:
            id_suffix = ""
            
        return '%s%s%s' % (url_prefix, self.pluralize(), id_suffix)   
    
    @classmethod
    def from_id(cls, course_id, identifier):
        url = "courses/%s/%s/%s" % (course_id, cls.pluralize(), identifier)
        return cls.from_uri(url)         
        
    @classmethod
    def all(cls, course_id):
        json = session.get("courses/%s/%s" % (course_id, cls.pluralize()))
        return [cls.from_json(o) for o in json[cls.pluralize()]]        





from sqlalchemy.ext.mutable import Mutable
from sqlalchemy.types import TypeDecorator, Unicode, DateTime
from datetime import datetime
import pytz
import json

# From 
#   http://stackoverflow.com/questions/2528189/can-sqlalchemy-datetime-objects-only-be-naive
# and
#   http://stackoverflow.com/questions/23316083/sqlalchemy-how-to-load-dates-with-timezone-utc-dates-stored-without-timezone

class UTCDateTime(TypeDecorator):

    impl = DateTime

    def process_bind_param(self, value, engine):
        if value is not None:
            return value.astimezone(pytz.utc)

    def process_result_value(self, value, engine):
        if value is not None:
            return datetime(value.year, value.month, value.day,
                            value.hour, value.minute, value.second,
                            value.microsecond, tzinfo=pytz.utc)

# From http://docs.sqlalchemy.org/en/rel_0_9/orm/extensions/mutable.html
class JSONEncodedDict(TypeDecorator):
    "Represents an immutable structure as a json-encoded string."

    impl = Unicode

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = unicode(json.dumps(value))
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value

class MutableDict(Mutable, dict):
    @classmethod
    def coerce(cls, key, value):
        "Convert plain dictionaries to MutableDict."

        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def __setitem__(self, key, value):
        "Detect dictionary set events and emit change events."

        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        "Detect dictionary del events and emit change events."

        dict.__delitem__(self, key)
        self.changed()
        
MutableDict.associate_with(JSONEncodedDict)        


def update_options(option_update_form, d):
    if len(option_update_form) > 0:
        for child_data in option_update_form:
            d[child_data.data["name"]] = child_data.data["value"]
        return True
    else:
        return False
    


from chisubmit.backend.webapp.api import db
from chisubmit.backend.webapp.api.assignments.models import Assignment,\
    GradeComponent
from chisubmit.backend.webapp.api.blueprints import api_endpoint
from flask import jsonify, request, abort
from chisubmit.backend.webapp.api.assignments.forms import UpdateAssignmentInput,\
    CreateAssignmentInput, RegisterAssignmentInput, SubmitAssignmentInput,\
    CancelSubmitAssignmentInput
from chisubmit.backend.webapp.auth.token import require_apikey
from chisubmit.backend.webapp.auth.authz import check_course_access_or_abort,\
    check_team_access_or_abort
from chisubmit.backend.webapp.api.courses.models import Course, CoursesStudents
from flask import g
from chisubmit.backend.webapp.api.users.models import User
from chisubmit.backend.webapp.api.teams.models import Team, StudentsTeams,\
    AssignmentsTeams
from chisubmit.common.utils import get_datetime_now_utc,\
    compute_extensions_needed
from sqlalchemy.sql.ddl import CreateTable


@api_endpoint.route('/courses/<course_id>/assignments', methods=['GET', 'POST'])
@require_apikey
def assignments(course_id):
    course = Course.query.filter_by(id=course_id).first()
    
    if course is None:
        abort(404)
        
    check_course_access_or_abort(g.user, course, 404)
    
    if request.method == 'GET':
        return jsonify(assignments=[assignment.to_dict()
                       for assignment in Assignment.query.filter_by(course_id=course_id).all()])

    check_course_access_or_abort(g.user, course, 404, roles=["instructor"])

    input_data = request.get_json(force=True)
    if not isinstance(input_data, dict):
        return jsonify(error='Request data must be a JSON Object'), 400
    form = CreateAssignmentInput.from_json(input_data)
    if not form.validate():
        return jsonify(errors=form.errors), 400

    assignment = Assignment()
    form.populate_obj(assignment)
    assignment.course_id = course_id
    db.session.add(assignment)
    db.session.commit()
        
    return jsonify({'assignment': assignment.to_dict()}), 201


@api_endpoint.route('/courses/<course_id>/assignments/<assignment_id>', methods=['PUT', 'GET'])
@require_apikey
def assignment(course_id, assignment_id):
    course = Course.query.filter_by(id=course_id).first()
    
    if course is None:
        abort(404)
        
    check_course_access_or_abort(g.user, course, 404)    
    
    assignment = Assignment.query.filter_by(id=assignment_id, course_id=course_id).first()
    # TODO 12DEC14: check permissions *before* 404
    if assignment is None:
        abort(404)

    if request.method == 'PUT':
        check_course_access_or_abort(g.user, course, 404, roles=["instructor"])
        
        input_data = request.get_json(force=True)
        if not isinstance(input_data, dict):
            return jsonify(error='Request data must be a JSON Object'), 400
        form = UpdateAssignmentInput.from_json(input_data)
        if not form.validate():
            return jsonify(errors=form.errors), 400

        assignment.set_columns(**form.patch_data)

        if 'grade_components' in form:
            for gc_data in form.grade_components.add:
                new_child = GradeComponent()
                gc_data.populate_obj(
                    type("", (), dict(new_child=new_child))(), 'new_child')
                new_child.course_id = course_id
                new_child.assignment_id = assignment_id
                db.session.add(new_child)
        db.session.commit()

    return jsonify({'assignment': assignment.to_dict()})


@api_endpoint.route('/courses/<course_id>/assignments/<assignment_id>/register', methods=['POST'])
@require_apikey
def assignment_register(course_id, assignment_id):
    course = Course.query.filter_by(id=course_id).first()
    
    if course is None:
        abort(404)
        
    check_course_access_or_abort(g.user, course, 404, roles=["student"])    
    
    assignment = Assignment.query.filter_by(id=assignment_id, course_id=course_id).first()
    # TODO 12DEC14: check permissions *before* 404
    if assignment is None:
        abort(404)

    if not g.user.is_student_in(course):
        error_msg = "You are not a student in course '%s'" % (course_id)
        return jsonify(errors={"register": error_msg}), 400

    if request.method == 'POST':       
        input_data = request.get_json(force=True)
        if not isinstance(input_data, dict):
            return jsonify(error='Request data must be a JSON Object'), 400
        form = RegisterAssignmentInput.from_json(input_data)
        if not form.validate():
            return jsonify(errors=form.errors), 400
        
        team_name = form.team_name.data
        if team_name is None or len(team_name) == 0:
            team_name = None

        partners = []
        for p in form.partners:
            user_id = p.data
            user = User.from_id(user_id)
            
            if user is None or not user.is_student_in(course):
                error_msg = "User '%s' is either not a valid user or not a student in course '%s'" % (user_id, course_id)
                return jsonify(errors={"partners": error_msg}), 400
            
            partners.append(user)

        students_in_team = [g.user] + partners
        
        warnings = []
        create_team = False
        
        teams = Team.find_teams_with_students(course_id, students_in_team)
                
        if len(teams) > 0:
            # Teams that have the assignment, but are *not* a perfect match
            have_assignment = []
            students_have_assignment = set()
            perfect_match = None
            
            for t in teams:
                if len(t.students) == len(students_in_team):
                    match = True
                    for s in students_in_team:
                        if s not in t.students:
                            match = False
                            break
                        
                    if match:
                        if perfect_match is None:
                            perfect_match = t
                        else:
                            # There shouldn't be more than one perfect match
                            error_msg = "There is more than one team with the exact same students in it." \
                                        "Please notify your instructor."  
                            return jsonify(errors={"fatal": error_msg}), 500
                    else:
                        if assignment in t.assignments:
                            have_assignment.append(t)
                            students_have_assignment.update(t.students)
                else:
                    if assignment in t.assignments:
                        have_assignment.append(t)
                        students_have_assignment.update(t.students)

            if len(have_assignment) > 0:
                error_msg = "'%s' is already registered for assignment '%s' in another team"
                error_msgs = [error_msg % (s.id, assignment.id) for s in students_have_assignment]
                return jsonify(errors={"partners": error_msgs}), 400                
                    
            if perfect_match is not None:
                if assignment in perfect_match.assignments:
                    user_st = [st for st in perfect_match.students_teams if st.student == g.user]
                    if len(user_st) != 1:
                        error_msg = "You appear to be registered twice in the same team." \
                                    "Please notify your instructor."
                        return jsonify(errors={"fatal": error_msg}), 500
                    user_st = user_st[0]
                    if user_st.status == StudentsTeams.STATUS_UNCONFIRMED:
                        user_st.status = StudentsTeams.STATUS_CONFIRMED
                        db.session.add(user_st)
                    elif user_st.status == StudentsTeams.STATUS_CONFIRMED:
                        warnings.append("You are already a confirmed member of this team")                        
                else:
                    at = AssignmentsTeams(team_id = perfect_match.id,
                                          course_id = course_id,
                                          assignment_id = assignment.id)
                    db.session.add(at)
            
                team_name = perfect_match.id
            else:
                create_team = True                        
        else:
            create_team = True
            
        if create_team:
            if team_name is None:
                team_name = "-".join(sorted([s.id for s in students_in_team]))
        
            team = Team(id = team_name, course_id=course_id)
            extension_policy = course.options.get("extension-policy", None)
            if extension_policy == "per_team":
                default_extensions = course.options.get("default-extensions", 0)
                team.extensions = default_extensions
            else:
                team.extensions = 0                
            db.session.add(team)
            
            st = StudentsTeams(team_id = team.id, 
                               course_id = course_id, 
                               student_id = g.user.id,
                               status = StudentsTeams.STATUS_CONFIRMED)
            db.session.add(st)
            
            for p in partners:
                st = StudentsTeams(team_id = team.id, 
                                   course_id = course_id, 
                                   student_id = p.id,
                                   status = StudentsTeams.STATUS_UNCONFIRMED)
                db.session.add(st)
                
            at = AssignmentsTeams(team_id = team.id,
                                  course_id = course_id,
                                  assignment_id = assignment.id)        
            db.session.add(at)
        
        db.session.commit()
            
        return jsonify({'team_name': team_name})


@api_endpoint.route('/courses/<course_id>/assignments/<assignment_id>/submit', methods=['POST'])
@require_apikey
def assignment_submit(course_id, assignment_id):
    now = get_datetime_now_utc()
    
    course = Course.query.filter_by(id=course_id).first()
    
    if course is None:
        abort(404)
        
    check_course_access_or_abort(g.user, course, 404, roles=["student"])    
    
    assignment = Assignment.query.filter_by(id=assignment_id, course_id=course_id).first()
    # TODO 12DEC14: check permissions *before* 404
    if assignment is None:
        abort(404)

    if not g.user.is_student_in(course):
        error_msg = "You are not a student in course '%s'" % (course_id)
        return jsonify(errors={"register": error_msg}), 400

    if request.method == 'POST':       
        input_data = request.get_json(force=True)
        if not isinstance(input_data, dict):
            return jsonify(error='Request data must be a JSON Object'), 400
        form = SubmitAssignmentInput.from_json(input_data)
        if not form.validate():
            return jsonify(errors=form.errors), 400
        
        team_id = form.team_id.data
        commit_sha = form.commit_sha.data
        extensions_requested = form.extensions.data
        dry_run = form.dry_run.data
        
        team = Team.query.filter_by(id=team_id).first()
        if team is None:
            abort(404)

        check_team_access_or_abort(g.user, team, 404)
        
        team_assignment = AssignmentsTeams.from_id(course.id,team.id,assignment.id)
        if team_assignment is None:
            msg = "Team '%s' is not registered for assignment '%s'" % (team.id, assignment.id)
            return jsonify(errors={"team":msg}), 400
        
        if team_assignment.is_ready_for_grading():
            msg = "You cannot re-submit assignment %s." % (assignment.id)
            msg = " You made a submission before the deadline, and the deadline has passed."
            return jsonify(errors={"submit":msg}), 400
        
        response = {}

        response["dry_run"] = dry_run
        
        response["prior_submission"] = {}
        response["prior_submission"]["submitted_at"] = team_assignment.submitted_at
        response["prior_submission"]["commit_sha"] = team_assignment.commit_sha
        response["prior_submission"]["extensions_used"] = team_assignment.extensions_used
        
        response["submission"] = {}
        response["submission"]["course_id"] = course.id
        response["submission"]["assignment_id"] = assignment.id
        response["submission"]["submitted_at"] = now
        response["submission"]["deadline"] = assignment.deadline
        
        extensions_needed = compute_extensions_needed(submission_time = now, deadline = assignment.deadline)
                
        extension_policy = course.options.get("extension-policy", None)
        extensions_available = team.get_extensions_available(extension_policy)
                
        if extensions_available < 0:
            error_msg = "The number of available extensions is negative"
            return jsonify(errors={"fatal": error_msg}), 500            

        response["submission"]["extensions_requested"] = extensions_requested
        response["submission"]["extensions_needed"] = extensions_needed

        response["team"] = {}
        response["team"]["id"] = team.id
        response["team"]["extensions_available_before"] = extensions_available
        
        if extensions_available + team_assignment.extensions_used >= extensions_needed and extensions_requested == extensions_needed: 
            response["success"] = True
            
            # If the team has already used extensions for a previous submission,
            # they don't count towards the number of extensions needed
            # They are 'credited' to the available extensions
            extensions_available += team_assignment.extensions_used
            
            extensions_available -= extensions_needed
            if not dry_run:
                team_assignment.extensions_used = extensions_needed
                team_assignment.commit_sha = commit_sha
                team_assignment.submitted_at = now
                
                db.session.add(team_assignment)
                db.session.commit()
        else:
            response["success"] = False
                
        response["team"]["extensions_available"] = extensions_available
        
        return jsonify(response)
        
        
@api_endpoint.route('/courses/<course_id>/assignments/<assignment_id>/cancel', methods=['POST'])
@require_apikey
def assignment_cancel_submit(course_id, assignment_id):
    course = Course.query.filter_by(id=course_id).first()
    
    if course is None:
        abort(404)
        
    check_course_access_or_abort(g.user, course, 404, roles=["student"])    
    
    assignment = Assignment.query.filter_by(id=assignment_id, course_id=course_id).first()
    # TODO 12DEC14: check permissions *before* 404
    if assignment is None:
        abort(404)

    if not g.user.is_student_in(course):
        error_msg = "You are not a student in course '%s'" % (course_id)
        return jsonify(errors={"register": error_msg}), 400

    if request.method == 'POST':       
        input_data = request.get_json(force=True)
        if not isinstance(input_data, dict):
            return jsonify(error='Request data must be a JSON Object'), 400
        form = CancelSubmitAssignmentInput.from_json(input_data)
        if not form.validate():
            return jsonify(errors=form.errors), 400
        
        team_id = form.team_id.data
        
        team = Team.query.filter_by(id=team_id).first()
        if team is None:
            abort(404)

        check_team_access_or_abort(g.user, team, 404)
        
        team_assignment = AssignmentsTeams.from_id(course.id,team.id,assignment.id)
        if team_assignment is None:
            msg = "Team '%s' is not registered for assignment '%s'" % (team.id, assignment.id)
            return jsonify(errors={"team":msg}), 400
        
        if team_assignment.is_ready_for_grading():
            msg = "You cannot cancel this submission for assignment %s." % (assignment.id)
            msg = " You made a submission before the deadline, and the deadline has passed."
            return jsonify(errors={"submit":msg}), 400        

        team_assignment.extensions_used = 0
        team_assignment.commit_sha = None
        team_assignment.submitted_at = None

        db.session.add(team_assignment)
        db.session.commit()
        
        return jsonify(success=True)
        



import abc
from functools import wraps
from flask import request, Response
from flask.globals import g

# Some parts from http://flask.pocoo.org/snippets/8/

class Auth(object):
    
    def __init__(self, server):
        self.server = server
        
    @abc.abstractmethod
    def check_auth(self, username, password):
        pass

auth_obj = None

def set_auth(obj):
    global auth_obj
    auth_obj = obj
    
def authenticate():
    global auth_obj

    """Sends a 401 response that enables basic auth"""
    if auth_obj is None:
        return Response(
        'The chisubmit server does not have an authentication.\n'
        'mechanism configured. Please contact the server administrator', 500)
    else:
        return Response(
        'Could not verify your access level for that URL.\n'
        'You have to login with proper credentials', 401,
        {'WWW-Authenticate': 'Basic realm="Login Required"'})        

def require_auth(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        from chisubmit.backend.webapp.api.users.models import User

        global auth_obj

        auth = request.authorization
        user = User.query.filter_by(id=unicode(auth.username)).first()

        if user is None:
            return authenticate()
        else:
            if not auth or not auth_obj or not auth_obj.check_auth(auth.username, auth.password):
                return authenticate()
            else:
                g.user = user
                return f(*args, **kwargs)
        
    return decorated    


from chisubmit.tests.common import ChisubmitMultiTestCase
from chisubmit.client.course import Course
from chisubmit.tests.fixtures import complete_course
import unittest
        
        
class CompleteCourse(ChisubmitMultiTestCase):
    
    FIXTURE = complete_course
        
    def test_get_course(self):
        
        self.get_test_client({"id":"admin", "api_key":"admin"})
        
        course = Course.from_id("cmsc40100")
        self.assertEquals(course.name, "Introduction to Software Testing")
        
        students = course.students
        
        self.assertEquals(len(students), 4)

            
        
                



#  Copyright (c) 2013-2014, The University of Chicago
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  - Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
#  - Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  - Neither the name of The University of Chicago nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.

import click

from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.client.user import User

@click.group(name="user")
@click.pass_context
def admin_user(ctx):
    pass

@click.command(name="add")
@click.argument('user_id', type=str)
@click.argument('first_name', type=str)
@click.argument('last_name', type=str)
@click.argument('email', type=str)
@click.pass_context
def admin_user_add(ctx, user_id, first_name, last_name, email):
    user = User.from_id(user_id)
    
    if user is not None:
        print "User %s already exists." % user_id
        ctx.exit(CHISUBMIT_FAIL)    
    
    user = User(id = user_id,
                first_name = first_name,
                last_name = last_name,
                email = email)

    return CHISUBMIT_SUCCESS

@click.command(name="remove")
@click.argument('user_id', type=str)
@click.pass_context
def admin_user_remove(ctx, user_id):
    # TODO
    print "NOT IMPLEMENTED"
    
    return CHISUBMIT_SUCCESS

admin_user.add_command(admin_user_add)
admin_user.add_command(admin_user_remove)






from chisubmit.tests.common import cli_test, ChisubmitIntegrationTestCase
from chisubmit.common.utils import get_datetime_now_utc
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL

from datetime import timedelta

class CLICompleteWorkflowExtensionsPerStudent(ChisubmitIntegrationTestCase):
            
    @cli_test
    def test_complete_with_extensions_per_student(self, runner):
        from chisubmit.backend.webapp.api.teams.models import Team
        from chisubmit.backend.webapp.api.courses.models import CoursesStudents

        course_id = u"cmsc40300"
        course_name = u"Foobarmentals of Foobar II"

        admin_id = u"admin"
        instructor_ids = [u"instructor"]
        grader_ids= [u"grader"]
        student_ids = [u"student1", u"student2", u"student3", u"student4"]
        
        all_users = instructor_ids + grader_ids + student_ids
        
        admin, instructors, graders, students = self.create_clients(runner, course_id, admin_id, instructor_ids, grader_ids, student_ids)
        self.create_users(admin, all_users)
        
        self.create_course(admin, course_id, course_name)

        result = admin.run("admin course set-option %s default-extensions 3" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        result = admin.run("admin course set-option %s extension-policy per_student" % (course_id))
        self.assertEquals(result.exit_code, 0)
        
        self.add_users_to_course(admin, course_id, instructors, graders, students)
        
        deadline = get_datetime_now_utc() - timedelta(hours=23)
        deadline = deadline.replace(tzinfo=None).isoformat(sep=" ")        

        result = instructors[0].run("instructor assignment add", 
                                    ["pa1", "Programming Assignment 1", deadline])
        self.assertEquals(result.exit_code, 0)

        deadline = get_datetime_now_utc() - timedelta(hours=47)
        deadline = deadline.isoformat(sep=" ")

        result = instructors[0].run("instructor assignment add", 
                                    ["pa2", "Programming Assignment 2", deadline])
        self.assertEquals(result.exit_code, 0)

        deadline = get_datetime_now_utc() - timedelta(hours=47)
        deadline = deadline.isoformat(sep=" ")

        result = instructors[0].run("instructor assignment add", 
                                    ["pa3", "Programming Assignment 3", deadline])
        self.assertEquals(result.exit_code, 0)

        deadline = get_datetime_now_utc() - timedelta(hours=23)
        deadline = deadline.replace(tzinfo=None).isoformat(sep=" ")        

        result = instructors[0].run("instructor assignment add", 
                                    ["pa4", "Programming Assignment 4", deadline])
        self.assertEquals(result.exit_code, 0)

        deadline = get_datetime_now_utc() + timedelta(hours=2)
        deadline = deadline.replace(tzinfo=None).isoformat(sep=" ")        

        result = instructors[0].run("instructor assignment add", 
                                    ["pa5", "Programming Assignment 5", deadline])
        self.assertEquals(result.exit_code, 0)
        
        
        teams = [u"the-flaming-foobars", 
                 u"the-magnificent-mallocs",
                 u"the-reticent-reallocs",
                 u"the-panicked-printfs"]        

        students_team = [ (students[0], students[1]),
                          (students[2], students[3]),
                          (students[0], students[2]),
                          (students[1], students[3])]
        

        self.register_team(students_team[0], teams[0], "pa1", course_id)
        self.register_team(students_team[1], teams[1], "pa2", course_id)
        
        team_git_paths, team_git_repos, team_commits = self.create_team_repos(admin, course_id, teams[0:2], students_team[0:2])

        self.register_team(students_team[2], teams[2], "pa3", course_id)
        self.register_team(students_team[2], teams[2], "pa4", course_id)
        self.register_team(students_team[2], teams[2], "pa5", course_id)
        self.register_team(students_team[3], teams[3], "pa3", course_id)
        self.register_team(students_team[3], teams[3], "pa4", course_id)
        self.register_team(students_team[3], teams[3], "pa5", course_id)
        
        x, y, z = self.create_team_repos(admin, course_id, teams[2:4], students_team[2:4])
        
        team_git_paths += x
        team_git_repos += y
        team_commits += z
        
        # Team 0 submits with one extension to pa1
        # Student 0 and 1 now have 2 extensions left each
        result = students_team[0][0].run("student assignment submit", 
                                         [teams[0], "pa1", team_commits[0][0].hexsha, 
                                          "--extensions", "1",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)
        
        cs1 =  CoursesStudents.from_id(course_id, students_team[0][0].user_id)
        cs2 =  CoursesStudents.from_id(course_id, students_team[0][1].user_id)
        t = Team.from_id(course_id, teams[0])
        
        self.assertEqual(t.get_extensions_available("per_student"), 2)
        self.assertEqual(cs1.get_extensions_available(), 2)
        self.assertEqual(cs2.get_extensions_available(), 2)
        
        # Team 1 submits with two extensions to pa2
        # Student 2 and 3 now have 1 extension left each
        result = students_team[1][0].run("student assignment submit", 
                                         [teams[1], "pa2", team_commits[1][0].hexsha, 
                                          "--extensions", "2",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        cs1 =  CoursesStudents.from_id(course_id, students_team[1][0].user_id)
        cs2 =  CoursesStudents.from_id(course_id, students_team[1][1].user_id)
        t = Team.from_id(course_id, teams[1])
        
        self.assertEqual(t.get_extensions_available("per_student"), 1)
        self.assertEqual(cs1.get_extensions_available(), 1)
        self.assertEqual(cs2.get_extensions_available(), 1)

        result = students_team[0][0].run("student team show", [teams[0]])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        result = students_team[1][0].run("student team show", [teams[1]])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)


        # Now we have two new teams:
        #
        # Team 2 with Student 0 and 2, with 2 and 1 extensions left respectively.
        # This team can only submit to projects requiring one extension
        # 
        # Team 3 with Student 1 and 3, with 2 and 1 extensions left respectively.
        # This team can only submit to projects requiring one extension

        # Team 2 tries to submit to pa3 with one extension (one less than needed)
        # and is denied
        result = students_team[2][0].run("student assignment submit", 
                                         [teams[2], "pa3", team_commits[2][0].hexsha, 
                                          "--extensions", "1",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_FAIL)

        cs1 =  CoursesStudents.from_id(course_id, students_team[2][0].user_id)
        cs2 =  CoursesStudents.from_id(course_id, students_team[2][1].user_id)
        t = Team.from_id(course_id, teams[2])
        
        self.assertEqual(t.get_extensions_available("per_student"), 1)
        self.assertEqual(cs1.get_extensions_available(), 2)
        self.assertEqual(cs2.get_extensions_available(), 1)


        # Team 2 tries to submit to pa3 with two extensions (the right number,
        # but not enough because one of the students only has one extension left)
        # and is denied
        result = students_team[2][0].run("student assignment submit", 
                                         [teams[2], "pa3", team_commits[2][0].hexsha, 
                                          "--extensions", "2",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_FAIL)

        cs1 =  CoursesStudents.from_id(course_id, students_team[2][0].user_id)
        cs2 =  CoursesStudents.from_id(course_id, students_team[2][1].user_id)
        t = Team.from_id(course_id, teams[2])
        
        self.assertEqual(t.get_extensions_available("per_student"), 1)
        self.assertEqual(cs1.get_extensions_available(), 2)
        self.assertEqual(cs2.get_extensions_available(), 1)


        # Team 2 tries submitting to pa4 with zero extensions (one less than needed)
        # and is denied
        result = students_team[2][0].run("student assignment submit", 
                                         [teams[2], "pa4", team_commits[2][1].hexsha, 
                                          "--extensions", "0",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_FAIL)

        cs1 =  CoursesStudents.from_id(course_id, students_team[2][0].user_id)
        cs2 =  CoursesStudents.from_id(course_id, students_team[2][1].user_id)
        t = Team.from_id(course_id, teams[2])
        
        self.assertEqual(t.get_extensions_available("per_student"), 1)
        self.assertEqual(cs1.get_extensions_available(), 2)
        self.assertEqual(cs2.get_extensions_available(), 1)

        # Team 2 tries submitting to pa4 with one extension (the correct number)
        # and it goes through
        result = students_team[2][0].run("student assignment submit", 
                                         [teams[2], "pa4", team_commits[2][1].hexsha, 
                                          "--extensions", "1",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)
                        
        cs1 =  CoursesStudents.from_id(course_id, students_team[2][0].user_id)
        cs2 =  CoursesStudents.from_id(course_id, students_team[2][1].user_id)
        t = Team.from_id(course_id, teams[2])
        
        self.assertEqual(t.get_extensions_available("per_student"), 0)
        self.assertEqual(cs1.get_extensions_available(), 1)
        self.assertEqual(cs2.get_extensions_available(), 0)


        # Team 2 tries submitting to pa5 with zero extensions and is accepted
        result = students_team[2][0].run("student assignment submit", 
                                         [teams[2], "pa5", team_commits[2][1].hexsha, 
                                          "--extensions", "0",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)
                        
        cs1 =  CoursesStudents.from_id(course_id, students_team[2][0].user_id)
        cs2 =  CoursesStudents.from_id(course_id, students_team[2][1].user_id)
        t = Team.from_id(course_id, teams[2])
        
        self.assertEqual(t.get_extensions_available("per_student"), 0)
        self.assertEqual(cs1.get_extensions_available(), 1)
        self.assertEqual(cs2.get_extensions_available(), 0)


        # Team 3 tries submitting to pa5 with zero extensions and is accepted
        result = students_team[3][0].run("student assignment submit", 
                                         [teams[3], "pa5", team_commits[3][1].hexsha, 
                                          "--extensions", "0",
                                          "--yes"])
        self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)
                        
        cs1 =  CoursesStudents.from_id(course_id, students_team[3][0].user_id)
        cs2 =  CoursesStudents.from_id(course_id, students_team[3][1].user_id)
        t = Team.from_id(course_id, teams[3])
        
        self.assertEqual(t.get_extensions_available("per_student"), 1)
        self.assertEqual(cs1.get_extensions_available(), 2)
        self.assertEqual(cs2.get_extensions_available(), 1)


        for team, student_team in zip(teams, students_team):
            result = student_team[0].run("student team show", [team])
            self.assertEquals(result.exit_code, CHISUBMIT_SUCCESS)

        
        



from chisubmit.repos import ConnectionString
from chisubmit.common import ChisubmitException
from chisubmit.repos.github import GitHubConnection
from chisubmit.repos.gitlab import GitLabConnection
from chisubmit.repos.testing import TestingConnection


class RemoteRepositoryConnectionFactory(object):

    server_types = {}

    @staticmethod
    def register_server_type(name, conn_cls):
        RemoteRepositoryConnectionFactory.server_types[name] = conn_cls

    @staticmethod
    def create_connection(connection_string, staging):
        cs = ConnectionString(connection_string)

        if not RemoteRepositoryConnectionFactory.server_types.has_key(cs.server_type):
            raise ChisubmitException("Unsupported server type in connection string: %s (expected one of: %s)" %
                                     (cs.server_type, ", ".join(RemoteRepositoryConnectionFactory.server_types.keys())))

        conn_cls = RemoteRepositoryConnectionFactory.server_types[cs.server_type]

        return conn_cls(cs, staging)

RemoteRepositoryConnectionFactory.register_server_type("GitHub", GitHubConnection)
RemoteRepositoryConnectionFactory.register_server_type("GitLab", GitLabConnection)
RemoteRepositoryConnectionFactory.register_server_type("Testing", TestingConnection)



#!/usr/bin/python

import sys
import random
import string
import requests
import click
import csv
from pprint import pprint as pp

def print_http_response(response):
    print "HTTP Status Code: %i" % response.status_code
    print
    print "HTTP Response"
    print "-------------"
    pp(response.headers.items())
    print
    pp(response.text)


@click.command(name="create-gitlab-users")
@click.argument('gitlab_hostname', type=str)
@click.argument('gitlab_token', type=str)
@click.argument('extern_uid_template', type=str)
@click.argument('csv_file', type=click.File('rb'))
@click.argument('csv_userid_column', type=str)
@click.argument('csv_fname_column', type=str)
@click.argument('csv_lname_column', type=str)
@click.argument('csv_email_column', type=str)
@click.option('--dry-run', is_flag=True)
@click.option('--id-from-email', is_flag=True)
def create_gitlab_users(gitlab_hostname, gitlab_token, extern_uid_template, csv_file, csv_userid_column, csv_fname_column, csv_lname_column, csv_email_column, dry_run, id_from_email):   
    headers = {"PRIVATE-TOKEN": gitlab_token}
    
    csvf = csv.DictReader(csv_file)
            
    for col in (csv_userid_column, csv_fname_column, csv_lname_column, csv_email_column):
        if col not in csvf.fieldnames:
            print "CSV file %s does not have a '%s' column" % (csv_file, col)
            ctx.exit(CHISUBMIT_FAIL)
        
    for entry in csvf:
        user_id = entry[csv_userid_column]
        if id_from_email:
            user_id = user_id.split("@")[0].strip()
        
        first_name = entry[csv_fname_column]
        last_name = entry[csv_lname_column]
        email = entry[csv_email_column]
    
        # Password doesn't actually matter since we use
        # LDAP authentication. Just in case, we set it to
        # something complicated
        passwd = ''.join([random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for i in range(25)])
    
        data = {"email": email, 
                "password": passwd,
                "username": user_id,
                "name": first_name + " " + last_name,
                "provider": "ldap",
                "extern_uid": extern_uid_template.replace("USER", user_id)}
    
        if dry_run:
            pp(data)
            print
        else:
            try:
                response = requests.post("https://%s/api/v3/users" % gitlab_hostname, 
                                         data=data,
                                         verify=False,
                                         headers=headers)
            except Exception, e:
                print "[ERROR] Unexpected exception when creating user %s" % user_id
        
            if response.status_code == 201:
                print "[OK] Created user %s" % user_id
            elif response.status_code == 404:
                print "[SKIP] User '%s' already exists" % user_id
            else:
                print "[ERROR] Unexpected error"
                print
                print_http_response(response)
                exit(1)

if __name__ == '__main__':
    create_gitlab_users()



from chisubmit.backend.webapp.api.courses import views
from chisubmit.backend.webapp.api.users import views
from chisubmit.backend.webapp.api.assignments import views
from chisubmit.backend.webapp.api.teams import views



from flask import jsonify, request, abort, g
from chisubmit.backend.webapp.api import db
from chisubmit.backend.webapp.api.blueprints import api_endpoint
from chisubmit.backend.webapp.api.courses.models import Course, CoursesInstructors,\
    CoursesStudents, CoursesGraders
from chisubmit.backend.webapp.api.assignments.models import Assignment
from chisubmit.backend.webapp.api.courses.forms import UpdateCourseInput, CreateCourseInput,\
    UpdateStudentInput
from chisubmit.backend.webapp.auth.token import require_apikey
from chisubmit.backend.webapp.auth.authz import check_course_access_or_abort,\
    check_admin_access_or_abort, check_user_ids_equal_or_abort
from chisubmit.backend.webapp.api.types import update_options

@api_endpoint.route('/courses', methods=['GET', 'POST'])
@require_apikey
def courses():
    if request.method == 'GET':
        # TODO: SQLAlchemy-fy this
        courses = Course.query.all()

        if not g.user.admin:
            courses = [c.to_dict() for c in courses if g.user.is_in_course(c)]
        else:
            courses = [c.to_dict() for c in courses]
        
        return jsonify(courses = courses), 200

    check_admin_access_or_abort(g.user, 404)

    input_data = request.get_json(force=True)
    if not isinstance(input_data, dict):
        return jsonify(error='Request data must be a JSON Object'), 400

    form = CreateCourseInput.from_json(input_data)
    if not form.validate():
        return jsonify(errors=form.errors), 400

    course = Course()
    form.populate_obj(course)
    db.session.add(course)
    db.session.commit()

    return jsonify({'course': course.to_dict()}), 201


@api_endpoint.route('/courses/<course_id>', methods=['PUT', 'GET'])
@require_apikey
def course(course_id):
    course = Course.query.filter_by(id=course_id).first()
    if course is None:
        abort(404)

    check_course_access_or_abort(g.user, course, 404)

    if request.method == 'GET':
        fields = []
        
        if g.user.admin:
            fields.append("options")
        
        if g.user.admin or g.user.is_instructor_in(course) or g.user.is_grader_in(course):
            fields += ['teams', 'courses_graders', 'courses_students', 'courses_instructors']

        return jsonify({'course': course.to_dict(show=fields)})
    
    if request.method == 'PUT':
        input_data = request.get_json(force=True)

        if not isinstance(input_data, dict):
            return jsonify(error='Request data must be a JSON Object'), 400
        form = UpdateCourseInput.from_json(input_data)
        if not form.validate():
            return jsonify(errors=form.errors), 400

        if g.user.is_instructor_in(course):
            course.set_columns(**form.patch_data)

        if 'assignments' in form:
            for assignment_data in form.assignments.add:
                check_course_access_or_abort(g.user, course, 403, roles=["instructor"])
                new_assignment = Assignment()
                # anonymous class to fool the populate_obj() call
                anonymous_class = type("", (), dict(assignment=new_assignment))()
                assignment_data.populate_obj(anonymous_class, 'assignment')
                db.session.add(new_assignment)

        if 'instructors' in form:
            for child_data in form.instructors.add:
                check_admin_access_or_abort(g.user, 403)
                new_child = CoursesInstructors()
                anonymous_class = type("", (), dict(child=new_child))()
                child_data.populate_obj(anonymous_class, 'child')
                
                git_usernames = course.options.get("git-usernames", None)

                if git_usernames == "user_id":
                    new_child.repo_info = {"git-username": new_child.instructor_id}
                
                db.session.add(new_child)
                
            for child_data in form.instructors.update:
                instructor_id = child_data["instructor_id"].data
                
                if instructor_id is None:
                    instructor_id = g.user.id
                
                if not g.user.admin:
                    check_user_ids_equal_or_abort(g.user.id, instructor_id, 403)
                
                course_instructor = CoursesInstructors.query.filter_by(
                        course_id=course_id).filter_by(
                        instructor_id=instructor_id).first()
                
                # TODO: A more descriptive 400 would be better
                if not course_instructor:
                    abort(404)
            
                if g.user.admin or g.user.is_instructor_in(course):
                    course_instructor.set_columns(**child_data.patch_data)
                    
                update_options(child_data.repo_info, course_instructor.repo_info)
                
                db.session.add(course_instructor)                

        if 'students' in form:
            for child_data in form.students.add:
                check_admin_access_or_abort(g.user, 403)
                new_child = CoursesStudents()
                anonymous_class = type("", (), dict(child=new_child))()
                child_data.populate_obj(anonymous_class, 'child')

                git_usernames = course.options.get("git-usernames", None)
                
                if git_usernames == "user_id":
                    new_child.repo_info = {"git-username": new_child.student_id}
                    
                extension_policy = course.options.get("extension-policy", None)
                if extension_policy == "per_student":
                    default_extensions = course.options.get("default-extensions", 0)
                    new_child.extensions = default_extensions
                else:
                    new_child.extensions = 0
                    
                db.session.add(new_child)
                
            for child_data in form.students.update:
                student_id = child_data["student_id"].data
                if student_id is None:
                    student_id = g.user.id                
                
                if not (g.user.admin or g.user.is_instructor_in(course)):
                    check_user_ids_equal_or_abort(g.user.id, student_id, 403)
                                    
                course_student = CoursesStudents.query.filter_by(
                        course_id=course_id).filter_by(
                        student_id=student_id).first()
                
                # TODO: A more descriptive 400 would be better
                if not course_student:
                    abort(404)

                if g.user.admin or g.user.is_instructor_in(course):
                    course_student.set_columns(**child_data.patch_data)
                update_options(child_data.repo_info, course_student.repo_info)
                
                db.session.add(course_student)
        
        if 'graders' in form:
            for child_data in form.graders.add:
                check_admin_access_or_abort(g.user, 403)
                new_child = CoursesGraders()
                anonymous_class = type("", (), dict(child=new_child))()
                child_data.populate_obj(anonymous_class, 'child')

                git_usernames = course.options.get("git-usernames", None)
                
                if git_usernames == "user_id":
                    new_child.repo_info = {"git-username": new_child.grader_id}
                
                db.session.add(new_child)

            for child_data in form.graders.update:
                grader_id = child_data["grader_id"].data

                if grader_id is None:
                    grader_id = g.user.id                
                
                if not (g.user.admin or g.user.is_instructor_in(course)):
                    check_user_ids_equal_or_abort(g.user.id, grader_id, 403)                
                
                course_grader = CoursesGraders.query.filter_by(
                        course_id=course_id).filter_by(
                        grader_id=grader_id).first()
                
                # TODO: A more descriptive 400 would be better
                if not course_grader:
                    abort(404)
            
                if g.user.admin or g.user.is_instructor_in(course):
                    course_grader.set_columns(**child_data.patch_data)
                update_options(child_data.repo_info, course_grader.repo_info)
                
                db.session.add(course_grader)
                
        if 'options' in form:
            if len(form.options.update) > 0:
                check_admin_access_or_abort(g.user, 403)                
                update_options(form.options.update, course.options)
                db.session.add(course)

        db.session.commit()

        return jsonify({'course': course.to_dict()})


@api_endpoint.route('/courses/<course_id>/students/<student_id>',
                    methods=['GET'])
@require_apikey
def course_student(course_id, student_id):
    course = Course.query.filter_by(id=course_id).first()
    if course is None:
        abort(404)

    check_course_access_or_abort(g.user, course, 404, roles=["instructor","grader"])    
    
    course_student = CoursesStudents.query.filter_by(
        course_id=course_id).filter_by(
        student_id=student_id).first()

    if not course_student:
        abort(404)

    return jsonify({'student': course_student.student.to_dict()}), 201




import click
from chisubmit.common import CHISUBMIT_SUCCESS, CHISUBMIT_FAIL
from chisubmit.client.course import Course
from chisubmit.cli.common import pass_course
from chisubmit.repos.factory import RemoteRepositoryConnectionFactory


@click.command(name="list")
@click.pass_context
def shared_course_list(ctx):
    courses = Course.all()
    
    for course in courses:
        print course.id, course.name

    return CHISUBMIT_SUCCESS

@click.command(name="set-default")
@click.argument('course_id', type=str)
@click.pass_context
def shared_course_set_default(ctx, course_id):
    ctx.obj['config']['default-course'] = course_id
    ctx.obj['config'].save()

@click.command(name="get-git-credentials")
@click.option('--username', prompt='Enter your git username')
@click.option('--password', prompt='Enter your git password', hide_input=True)
@click.option('--no-save', is_flag=True)
@click.option('--delete-permissions', is_flag=True)
@click.option('--staging', is_flag=True)
@pass_course
@click.pass_context
def shared_course_get_git_credentials(ctx, course, username, password, no_save, delete_permissions, staging):
    if not staging:
        connstr_field = "git-server-connstr"
    else:
        connstr_field = "git-staging-connstr"
    
    if not course.options.has_key(connstr_field):
        print "Course '%s' doesn't seem to be configured to use a Git server." % course.id
        ctx.exit(CHISUBMIT_FAIL)
        
    connstr = course.options[connstr_field]

    conn = RemoteRepositoryConnectionFactory.create_connection(connstr, staging = staging)
    server_type = conn.get_server_type_name()

    token, existing = conn.get_credentials(username, password, delete_repo = delete_permissions)

    if token is None:
        print "Unable to create token. Incorrect username/password."
    else:
        if not no_save:
            if ctx.obj['config']['git-credentials'] is None:
                ctx.obj['config']['git-credentials'] = {}
            ctx.obj['config']['git-credentials'][server_type] = token
            ctx.obj['config'].save()
        
        if existing:
            print "Your existing %s access token is: %s" % (server_type, token)
        else:
            print "The following %s access token has been created: %s" % (server_type, token)

        if not no_save:
            print "chisubmit has been configured to use this token from now on."

    return CHISUBMIT_SUCCESS
